import numpy as np
from dryml.data import NumpyDataset
from dryml.data import util
import dryml
import pytest


def test_data_util_slicer_1():
    array_shape = (10, 20)
    data1 = np.random.random(array_shape)
    data2 = np.random.random(array_shape)

    slice1 = slice(None, 5)
    slice2 = slice(5, None)

    data_tup = (data1, data2)

    data_tup_slice1 = util.nested_slice(data_tup, slice1)

    assert np.all(data_tup_slice1[0] == data1[slice1])
    assert np.all(data_tup_slice1[1] == data2[slice1])

    data_tup_slice2 = util.nested_slice(data_tup, slice2)

    assert np.all(data_tup_slice2[0] == data1[slice2])
    assert np.all(data_tup_slice2[1] == data2[slice2])


def test_data_util_slicer_2():
    array_shape = (10, 20)
    data1 = np.random.random(array_shape)
    data2 = np.random.random(array_shape)

    slice1 = slice(None, 5)
    slice2 = slice(5, None)

    data_dict = {'key1': data1, 'key2': data2}

    data_dict_slice1 = util.nested_slice(data_dict, slice1)

    assert np.all(data_dict_slice1['key1'] == data1[slice1])
    assert np.all(data_dict_slice1['key2'] == data2[slice1])

    data_dict_slice2 = util.nested_slice(data_dict, slice2)

    assert np.all(data_dict_slice2['key1'] == data1[slice2])
    assert np.all(data_dict_slice2['key2'] == data2[slice2])


def test_data_util_slicer_3():
    array_shape = (10, 20)
    data1 = np.random.random(array_shape)
    data2 = np.random.random(array_shape)
    data3 = np.random.random(array_shape)

    slice1 = slice(None, 5)
    slice2 = slice(5, None)

    data = (data1, {'key1': data2, 'key2': data3})

    data_slice1 = util.nested_slice(data, slice1)

    assert np.all(data_slice1[0] == data1[slice1])
    assert np.all(data_slice1[1]['key1'] == data2[slice1])
    assert np.all(data_slice1[1]['key2'] == data3[slice1])

    data_slice2 = util.nested_slice(data, slice2)

    assert np.all(data_slice2[0] == data1[slice2])
    assert np.all(data_slice2[1]['key1'] == data2[slice2])
    assert np.all(data_slice2[1]['key2'] == data3[slice2])


def test_data_util_slicer_4():
    array_shape = (10, 20)
    data1 = np.random.random(array_shape)
    data2 = np.random.random(array_shape)
    data3 = np.random.random(array_shape)

    slice1 = slice(None, 5)
    slice2 = slice(5, None)

    data = {'key1': (data1, data2), 'key2': data3}

    data_slice1 = util.nested_slice(data, slice1)

    assert np.all(data_slice1['key1'][0] == data1[slice1])
    assert np.all(data_slice1['key1'][1] == data2[slice1])
    assert np.all(data_slice1['key2'] == data3[slice1])

    data_slice2 = util.nested_slice(data, slice2)

    assert np.all(data_slice2['key1'][0] == data1[slice2])
    assert np.all(data_slice2['key1'][1] == data2[slice2])
    assert np.all(data_slice2['key2'] == data3[slice2])


def test_numpy_dataset_1():
    batch_size = 10

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data, batch_size=batch_size)

    # assert len(dataset) == batch_size

    i = 0
    for el in dataset.unbatch():
        assert np.all(el == data[i])
        i += 1


def test_numpy_dataset_2():
    batch_size = 10

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data)

    # assert len(dataset) == batch_size

    i = 0
    for el in dataset.unbatch():
        assert np.all(el == data[i])
        i += 1


def test_numpy_dataset_3():
    batch_size = 10

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data, batch_size=batch_size)

    # assert len(dataset) == batch_size

    v = next(iter(dataset))

    assert np.all(v == data)


def test_numpy_dataset_4():
    batch_size = 10

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data)

    # assert len(dataset) == batch_size

    v = next(iter(dataset))
    assert np.all(v == data)


def test_numpy_dataset_5():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows)

    # assert len(dataset) == batch_size

    i = 0
    for el in dataset:
        assert np.all(el == batch_rows[i])
        i += 1


def test_numpy_dataset_6():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows)

    # assert len(dataset) == batch_size

    v = next(iter(dataset.batch(batch_size=batch_size)))

    all_data = np.stack(batch_rows, axis=0)

    assert np.all(v == all_data)


def test_numpy_dataset_7():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows)

    # assert len(dataset) == batch_size

    sub_batch_size = batch_size//4

    dataset = dataset.batch(batch_size=sub_batch_size)

    # assert len(dataset) == batch_size

    i = 0
    for el in dataset:
        sub_batch = np.stack(
            batch_rows[i*sub_batch_size:(i+1)*sub_batch_size], axis=0)
        assert np.all(el == sub_batch)
        i += 1


def test_numpy_dataset_8():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows)

    # assert len(dataset) == batch_size

    dataset2 = dataset.batch(batch_size//4).unbatch()

    # assert len(dataset2) == batch_size

    for el, el2 in zip(dataset, dataset2):
        assert np.all(el == el2)


def test_numpy_dataset_9():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows)

    # assert len(dataset) == batch_size

    dataset = dataset.as_indexed()

    # assert len(dataset) == batch_size

    i = 0
    for idx, el in dataset:
        assert idx == i
        assert np.all(el == batch_rows[i])
        i += 1


def test_numpy_dataset_10():
    batch_size = 32

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data)

    # assert len(dataset) == batch_size

    dataset = dataset.as_indexed().unbatch()

    # assert len(dataset) == batch_size

    i = 0
    for idx, el in dataset:
        assert idx == i
        assert np.all(el == data[i])
        i += 1


def test_numpy_dataset_11():
    batch_size = 32

    data = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data)

    # assert len(dataset) == batch_size

    dataset = dataset.as_indexed().as_not_indexed().unbatch()

    # assert len(dataset) == batch_size

    i = 0
    for el in dataset:
        assert np.all(el == data[i])
        i += 1


def test_numpy_dataset_12():
    batch_size = 32

    data1 = np.random.random((batch_size, 20))
    data2 = np.random.random((batch_size, 20))

    dataset = NumpyDataset([(data1, data2)], batch_size=batch_size)

    # assert len(dataset) == batch_size

    dataset = dataset.as_indexed().unbatch()

    # assert len(dataset) == batch_size

    i = 0
    for idx, el in dataset:
        assert idx == i
        assert np.all(el[0] == data1[i])
        assert np.all(el[1] == data2[i])
        i += 1


def test_numpy_dataset_13():
    batch_size = 32

    data1 = np.random.random((batch_size, 20))
    data2 = np.random.random((batch_size, 20))

    dataset = NumpyDataset([(data1, data2)], batch_size=batch_size)

    # assert len(dataset) == batch_size

    dataset_2 = dataset.batch(batch_size=2)

    # assert len(dataset_2) == batch_size

    assert dataset is not dataset_2


def test_numpy_dataset_14():
    batch_size = 32

    batch_rows = []
    for i in range(batch_size):
        batch_rows.append(np.random.random((20,)))

    dataset = NumpyDataset(batch_rows).batch(batch_size=batch_size)

    # assert len(dataset) == batch_size

    assert dataset is dataset.batch(batch_size=batch_size)


def test_numpy_dataset_15():
    batch_size = 32
    data_block = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data_block, batch_size=batch_size)

    # assert len(dataset) == batch_size

    take_num = 10
    dataset = dataset.unbatch().take(take_num)

    # assert len(dataset) == take_num

    count = 0
    for el in dataset:
        count += 1
        assert el.shape == (20,)

    assert count == take_num


def test_numpy_dataset_16():
    batch_size = 32
    data_block = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data_block, batch_size=batch_size)

    # assert len(dataset) == batch_size

    el_a = dataset.peek()
    el_b = dataset.peek()

    assert np.all(el_a == el_b)


def test_numpy_dataset_17():
    batch_size = 32
    data_block = np.random.random((batch_size, 20))

    dataset = NumpyDataset(data_block, batch_size=batch_size)

    # assert len(dataset) == batch_size

    for el_a, el_b in zip(dataset, dataset):
        assert np.all(el_a == el_b)


def test_numpy_dataset_18():
    batch_size = 32
    data_block_x = np.random.random((batch_size, 20))
    data_block_y = np.random.random((batch_size, 20))

    dataset = NumpyDataset(
        (data_block_x, data_block_y),
        batch_size=batch_size,
        supervised=True)

    # assert len(dataset) == batch_size

    el_a = dataset.peek()
    el_b = dataset.peek()

    assert type(el_a) is tuple
    assert type(el_b) is tuple

    assert np.all(el_a[0] == el_b[0])
    assert np.all(el_a[1] == el_b[1])


def test_chain_transforms_9():
    batch_size = 32
    data_block = np.random.random((batch_size, 5))

    result = dryml.data.NumpyDataset(data_block) \
                       .batch() \
                       .apply_X(func=lambda X: X**2) \
                       .numpy() \
                       .peek()

    assert type(result) is np.ndarray
    assert np.all(data_block**2 == result)


# Define equality functions which will be needed
def np_eq(el1, el2):
    return np.all(el1 == el2)


def tf_eq(el1, el2):
    tf = pytest.importorskip('tensorflow')
    try:
        tf.debugging.assert_equal(el1, el2)
        return True
    except Exception:
        return False


def torch_eq(el1, el2):
    torch = pytest.importorskip('torch')
    try:
        torch.testing.assert_close(el1, el2, rtol=0, atol=0)
        return True
    except Exception:
        return False


datasets_to_test = []


def append_dataset_gen(f):
    datasets_to_test.append(f)
    return f


@append_dataset_gen
def np_dataset_1():
    dataset = np.random.random((50, 50))
    dataset = dryml.data.NumpyDataset(dataset)
    return dataset, np_eq


@append_dataset_gen
def np_dataset_2():
    dataset = np.random.random((50, 50))
    dataset = dryml.data.NumpyDataset(dataset).unbatch()
    return dataset, np_eq


@append_dataset_gen
def np_dataset_3():
    dataset = np.random.random((50, 50))
    dataset = dryml.data.NumpyDataset(dataset) \
                        .apply_X(lambda x: x**2)
    return dataset, np_eq


@append_dataset_gen
def tf_dataset_1():
    tf = pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    tf_data = tf.data.Dataset.from_tensor_slices(data)
    from dryml.data.tf import TFDataset
    dataset = TFDataset(tf_data)
    return dataset, tf_eq


@append_dataset_gen
def tf_dataset_2():
    tf = pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    tf_data = tf.data.Dataset.from_tensor_slices(data)
    from dryml.data.tf import TFDataset
    dataset = TFDataset(tf_data).unbatch()
    return dataset, tf_eq


@append_dataset_gen
def tf_dataset_3():
    tf = pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    tf_data = tf.data.Dataset.from_tensor_slices(data)
    from dryml.data.tf import TFDataset
    dataset = TFDataset(tf_data).apply_X(lambda x: tf.multiply(x, x))
    return dataset, tf_eq


@append_dataset_gen
def np_tf_dataset_1():
    pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    ds = dryml.data.NumpyDataset(data)
    ds = ds.tf()
    return ds, tf_eq


@append_dataset_gen
def np_tf_dataset_2():
    tf = pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    ds = dryml.data.NumpyDataset(data)
    ds = ds.tf().apply_X(
        lambda x: tf.multiply(x, x))
    return ds, tf_eq


@append_dataset_gen
def np_tf_dataset_3():
    tf = pytest.importorskip('tensorflow')
    data = np.random.random((50, 50))
    ds = dryml.data.NumpyDataset(data)

    def mult(x):
        return tf.multiply(x, x)

    ds = ds.tf().apply_X(mult)

    return ds, tf_eq


@append_dataset_gen
def torch_dataset_1():
    torch = pytest.importorskip('torch')
    data = np.random.random((50, 50))
    from dryml.data.torch import TorchDataset, TorchIterableDatasetWrapper
    torch_ds = TorchIterableDatasetWrapper(
        lambda: map(lambda x: torch.tensor(x), iter(data)))
    dataset = TorchDataset(torch_ds)
    return dataset, torch_eq


@append_dataset_gen
def torch_dataset_2():
    torch = pytest.importorskip('torch')
    data = np.random.random((50, 50))
    from dryml.data.torch import TorchDataset, TorchIterableDatasetWrapper
    torch_ds = TorchIterableDatasetWrapper(
         lambda: map(lambda x: torch.tensor(x), iter(data)))
    dataset = TorchDataset(torch_ds).unbatch()
    return dataset, torch_eq


@append_dataset_gen
def torch_dataset_3():
    torch = pytest.importorskip('torch')
    data = np.random.random((50, 50))
    from dryml.data.torch import TorchDataset, TorchIterableDatasetWrapper
    torch_ds = TorchIterableDatasetWrapper(
         lambda: map(lambda x: torch.tensor(x), iter(data)))
    dataset = TorchDataset(torch_ds).apply_X(lambda x: torch.multiply(x, x))
    return dataset, torch_eq


@pytest.mark.parametrize(
    'dataset_gen', datasets_to_test)
def test_double_collect_1(dataset_gen):
    dataset, eq_func = dataset_gen()
    collect_elements = dataset.collect()
    collect_elements_2 = dataset.collect()

    assert len(collect_elements) > 0
    assert len(collect_elements) == len(collect_elements_2)
    for el1, el2 in zip(collect_elements, collect_elements_2):
        assert eq_func(el1, el2)


@pytest.mark.parametrize(
    'dataset_gen', datasets_to_test)
def test_double_collect_2(dataset_gen):
    dataset, eq_func = dataset_gen()
    collect_elements = dataset.take(1).collect()
    collect_elements_2 = dataset.take(1).collect()

    assert len(collect_elements) > 0
    assert len(collect_elements) == len(collect_elements_2)
    for el1, el2 in zip(collect_elements, collect_elements_2):
        assert eq_func(el1, el2)


@pytest.mark.parametrize(
    'dataset_gen', datasets_to_test)
def test_shuffle_check_1(dataset_gen):
    dataset, _ = dataset_gen()
    orig_count = dataset.count()
    if dataset.batched:
        print("batched branch")
        batch_size = dataset.batch_size
        new_count = dataset.shuffle(10000).batch(batch_size=batch_size).count()
    else:
        print("unbatched branch")
        new_count = dataset.shuffle(10000).count()
    assert orig_count == new_count



import pytest
import dryml
import os
import tempfile

import objects


@pytest.mark.usefixtures("create_temp_dir")
def test_container_selector_1(create_temp_dir):
    # Create objects
    obj1 = objects.TestClassA(base_msg="Test1", item=5)
    obj2 = objects.TestClassB([1, 2, 3], base_msg="Test1")

    # Create containers for these objects
    obj1_cont = dryml.repo.RepoContainer.from_object(
        obj1, directory=create_temp_dir)
    obj2_cont = dryml.repo.RepoContainer.from_object(
        obj2, directory=create_temp_dir)

    # Create selector
    sel = dryml.Selector(cls=objects.TestClassA)

    # Test selectors work with built classes
    assert sel(obj1_cont.definition())
    assert not sel(obj2_cont.definition())


@pytest.mark.usefixtures("create_temp_dir")
def test_container_selector_2(create_temp_dir):
    # Create objects
    obj1 = objects.TestClassA(base_msg="Test1", item=5)
    obj2 = objects.TestClassB([1, 2, 3], base_msg="Test1")

    # Create containers for these objects
    obj1_cont = dryml.repo.RepoContainer.from_object(
        obj1, directory=create_temp_dir)
    obj2_cont = dryml.repo.RepoContainer.from_object(
        obj2, directory=create_temp_dir)

    # Save to disk, and unload these objects
    obj1_cont.save()
    obj2_cont.save()
    obj1_cont.unload()
    obj2_cont.unload()

    # Create selector
    sel = dryml.Selector(cls=objects.TestClassA)

    # Test selectors work with built classes
    assert sel(obj1_cont.definition())
    assert not sel(obj2_cont.definition())


@pytest.mark.usefixtures("create_temp_dir")
def test_add_retrieve_object_1(create_temp_dir):
    obj = objects.HelloStr(msg='test')

    repo = dryml.Repo(create_temp_dir)

    repo.add_object(obj)

    assert len(repo) == 1

    obj_ret = repo.get()

    assert type(obj_ret) is not list

    assert obj_ret.definition().get_individual_id() == \
        obj.definition().get_individual_id()


def test_add_retrieve_objects_2():
    repo = dryml.Repo()

    objs = []
    objs.append(objects.HelloStr(msg='test'))
    objs.append(objects.HelloInt(msg=10))
    objs.append(objects.HelloInt(msg=10))
    objs.append(objects.TestClassA(item=[10, 10]))
    objs.append(objects.TestClassB('test'))

    for obj in objs:
        repo.add_object(obj)

    assert type(repo.get(selector=dryml.Selector(
        cls=objects.HelloStr, kwargs={'msg': 'test'}))) is not list
    assert len(repo.get(selector=dryml.Selector(
        cls=objects.HelloInt, kwargs={'msg': 10}))) == 2
    assert type(repo.get(selector=dryml.Selector(
        cls=objects.TestClassA, kwargs={'item': [10, 10]}))) is not list
    assert type(repo.get(selector=dryml.Selector(
        cls=objects.TestClassB, args=['test']))) is not list


def test_add_retrieve_objects_3():
    """
    Should be able to add all objects within an object at once.
    """

    repo = dryml.Repo()

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))

    repo.add_object(obj)

    assert len(repo.get()) == 3
    assert type(repo.get(selector=dryml.Selector(
        cls=objects.TestNest, args=(10,)))) is not list
    assert type(repo.get(selector=dryml.Selector(
        cls=objects.HelloTrainableD))) is not list
    obj = repo.get(
        selector=dryml.Selector(
            cls=objects.TestNest,
            args=(dryml.Selector(objects.HelloTrainableD),),),
        sel_kwargs={'verbosity': 2})
    assert type(obj) is not list


def test_add_retrieve_objects_4():
    """
    Object hierarchy should work during selection
    """

    repo = dryml.Repo()

    parent_cls_obj = objects.TestBase()

    repo.add_object(parent_cls_obj)

    assert dryml.utils.count(repo.get(
        selector=dryml.Selector(objects.TestBase))) == 1
    try:
        repo.get(selector=dryml.Selector(objects.TestClassA))
        assert False
    except KeyError:
        pass


@pytest.mark.xfail
def test_try_write():
    repo = dryml.Repo()

    repo.add_object(objects.HelloStr(msg='test'))

    repo.save()


def test_get_api_1():
    repo = dryml.Repo()

    repo.add_object(objects.HelloStr(msg='test'))

    repo.get(sel_kwargs={'verbosity': 10})


def test_get_api_2():
    repo = dryml.Repo()

    obj1 = objects.TestNest(1)
    obj2 = objects.TestNest(2)
    obj3 = objects.TestNest(3)
    obj4 = objects.TestNest(4)

    repo.add_object(obj1)
    repo.add_object(obj2)
    repo.add_object(obj3)
    repo.add_object(obj4)

    # Get container for first object
    res1 = repo[obj1]
    assert type(res1) is objects.TestNest
    assert obj1.dry_id == res1.dry_id
    assert res1.A == 1

    res3 = repo[obj3]
    assert type(res3) is objects.TestNest
    assert res3.dry_id == obj3.dry_id
    assert res3.A == 3


@pytest.mark.usefixtures("create_temp_dir")
def test_get_api_3(create_temp_dir):
    repo = dryml.Repo()

    test_obj_def = dryml.ObjectDef(
        objects.HelloStr,
        msg='test')

    try:
        test_obj = repo.get(test_obj_def)
        assert False
    except KeyError:
        pass

    test_obj = repo.get(test_obj_def, build_missing_def=True)

    assert len(repo) == 1
    assert dryml.Selector.build(test_obj_def)(test_obj)


@pytest.mark.usefixtures("create_temp_dir")
def test_get_api_4(create_temp_dir):
    repo = dryml.Repo()

    test_obj_def = dryml.ObjectDef(
        objects.HelloStr,
        msg='test')

    try:
        repo.get(test_obj_def, build_missing_def=False)
        assert False
    except KeyError:
        pass

    assert len(repo) == 0


@pytest.mark.usefixtures("create_temp_dir")
def test_get_api_5(create_temp_dir):
    repo = dryml.Repo()

    test_obj_def = dryml.ObjectDef(
        objects.TestNest,
        dryml.ObjectDef(
            objects.TestNest2,
            A=5)
        )

    obj = repo.get(test_obj_def, build_missing_def=True)

    assert len(repo) == 2

    obj2 = repo.get(test_obj_def)
    assert obj is obj2
    assert len(repo) == 2


@pytest.mark.usefixtures("create_temp_dir")
def test_write_1(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    objs = []

    objs.append(objects.HelloStr(msg='test'))
    objs.append(objects.HelloInt(msg=10))
    objs.append(objects.HelloInt(msg=10))
    objs.append(objects.TestClassA(item=[10, 10]))
    objs.append(objects.TestClassB('test'))

    for obj in objs:
        repo.add_object(obj)

    repo.save()

    # Delete repo
    del repo

    repo = dryml.Repo(create_temp_dir)

    assert len(repo) == 5

    obj = repo.get(selector=dryml.Selector(
        cls=objects.HelloStr, kwargs={'msg': 'test'}))
    assert type(obj) is not list
    assert objs[0].definition().get_individual_id() == \
        obj.definition().get_individual_id()

    obj_list = repo.get(selector=dryml.Selector(
        cls=objects.HelloInt, kwargs={'msg': 10}))
    assert len(obj_list) == 2
    assert objs[1].definition().get_category_id() == \
        obj_list[0].definition().get_category_id()
    assert objs[1].definition().get_category_id() == \
        obj_list[1].definition().get_category_id()

    obj = repo.get(selector=dryml.Selector(
        cls=objects.TestClassA, kwargs={'item': [10, 10]}))
    assert type(obj) is not list
    assert objs[3].definition().get_individual_id() == \
        obj.definition().get_individual_id()

    obj = repo.get(selector=dryml.Selector(
        cls=objects.TestClassB, args=['test']))
    assert type(obj) is not list
    assert objs[4].definition().get_individual_id() == \
        obj.definition().get_individual_id()


@pytest.mark.usefixtures("create_temp_dir")
def test_reload_1(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    objs = []

    objs.append(objects.TestClassA(item=[10]))
    objs.append(objects.TestClassA(item=[10, 10]))
    objs.append(objects.TestClassA(item='a'))

    for obj in objs:
        repo.add_object(obj)

    repo.reload_objs(selector=dryml.Selector(cls=objects.TestClassA),
                     as_cls=objects.TestClassA2)

    obj = repo.get(selector=dryml.Selector(
        cls=objects.TestClassA2, kwargs={'item': [10]}))
    assert objs[0].dry_kwargs['item'] == obj.dry_kwargs['item']

    obj = repo.get(selector=dryml.Selector(
        cls=objects.TestClassA2, kwargs={'item': [10, 10]}))
    assert objs[1].dry_kwargs['item'] == obj.dry_kwargs['item']

    obj = repo.get(selector=dryml.Selector(
        cls=objects.TestClassA2, kwargs={'item': 'a'}))
    assert objs[2].dry_kwargs['item'] == obj.dry_kwargs['item']


@pytest.mark.usefixtures("create_temp_dir")
def test_save_1(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'))

    # Save objects in repository
    repo.save()

    # Delete the repo
    del repo

    # Load the repository objects should not be loaded right away
    repo = dryml.Repo(create_temp_dir)

    try:
        repo.get(only_loaded=True)
        assert False
    except KeyError:
        pass

    repo.save()

    assert len(os.listdir(create_temp_dir)) == 1


@pytest.mark.usefixtures("create_temp_dir")
def test_save_2(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'), filepath='test_file')

    # Save objects in repository
    repo.save()

    # Delete the repo
    del repo

    # Load the repository objects should not be loaded right away
    repo = dryml.Repo(create_temp_dir)

    try:
        repo.get(only_loaded=True)
        assert False
    except KeyError:
        pass

    repo.save()

    assert len(os.listdir(create_temp_dir)) == 1


@pytest.fixture
def prep_and_clean_test_dir2():
    with tempfile.TemporaryDirectory() as dir1, \
         tempfile.TemporaryDirectory() as dir2:
        yield dir1, dir2


def test_save_3(prep_and_clean_test_dir2):
    dir1, dir2 = prep_and_clean_test_dir2
    repo = dryml.Repo(dir1, create=True)

    repo.add_object(objects.HelloStr(msg='test'),
                    filepath='test_file')
    repo.add_object(objects.HelloInt(msg=5),
                    filepath=os.path.join(dir2, 'test_file'))

    # Save objects in repository
    repo.save()

    # Delete the repo
    del repo

    assert len(os.listdir(dir1)) == 1
    assert len(os.listdir(dir2)) == 1

    # Load the repository objects should not be loaded right away
    repo = dryml.Repo(dir1)

    try:
        repo.get(only_loaded=True)
        assert False
    except KeyError:
        pass

    del repo

    repo = dryml.Repo(dir2)

    try:
        repo.get(only_loaded=True)
        assert False
    except KeyError:
        pass


def test_save_4(prep_and_clean_test_dir2):
    dir1, dir2 = prep_and_clean_test_dir2
    repo = dryml.Repo(dir1, create=True)

    repo.add_object(objects.HelloStr(msg='test'), filepath='test_file')
    repo.add_object(objects.HelloInt(msg=5))

    # Save objects in repository
    repo.save()
    repo.save(directory=dir2)

    # Delete the repo
    del repo

    assert set(os.listdir(dir1)) == set(os.listdir(dir2))


@pytest.mark.usefixtures("create_temp_dir")
def test_save_5(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'), filepath='test_file')

    repo.save_objs_on_deletion = True

    # Delete the repo
    del repo

    assert len(os.listdir(create_temp_dir)) == 1


@pytest.mark.usefixtures("create_temp_dir")
def test_save_6(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'), filepath='test_file')

    repo.save_and_cache()

    try:
        repo.get(only_loaded=True)
        assert False
    except KeyError:
        pass
    assert len(os.listdir(create_temp_dir)) == 1


@pytest.mark.usefixtures("create_temp_dir")
def test_save_7(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    obj1 = objects.TestNest2(A=5)
    obj2 = objects.TestNest(obj1)

    repo.add_object(obj2)

    assert len(repo) == 2

    assert obj1 is repo[obj1]
    assert obj1 is repo[obj1.dry_id]
    assert obj2 is repo[obj2]
    assert obj2 is repo[obj2.dry_id]

    repo.save(obj2)

    assert len(os.listdir(create_temp_dir)) == 2

    del repo

    repo = dryml.Repo(create_temp_dir)

    assert len(repo) == 2

    assert type(repo.get(obj1)) is not list
    assert type(repo.get(obj2)) is not list


@pytest.mark.usefixtures("create_temp_dir")
def test_save_8(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    obj1 = objects.TestNest2(A=5)
    obj2 = objects.TestNest(obj1)

    repo.save(obj2)

    assert len(repo) == 2

    assert len(os.listdir(create_temp_dir)) == 2


@pytest.mark.usefixtures("create_temp_dir")
def test_delete_1(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'))

    repo.save()

    assert len(os.listdir(create_temp_dir)) == 1

    repo.delete()

    assert len(os.listdir(create_temp_dir)) == 0
    try:
        repo.get(load_objects=True)
        assert False
    except KeyError:
        pass


@pytest.mark.usefixtures("create_temp_dir")
def test_delete_2(create_temp_dir):
    repo = dryml.Repo(create_temp_dir, create=True)

    repo.add_object(objects.HelloStr(msg='test'))

    assert len(repo) == 1

    repo.delete()

    assert len(repo) == 0


@pytest.mark.usefixtures("create_temp_dir")
def test_object_save_restore_with_repo_1(create_temp_dir):
    """
    We test save and restore of nested objects through arguments
    """
    import objects

    repo = dryml.Repo(create_temp_dir, create=True)

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    # Add and save object in repo
    repo.add_object(data_obj1)
    repo.save()

    # Enclose them in another object
    obj = objects.TestClassC(data_obj1, B=data_obj1)

    # Load the object from the file
    obj2 = obj.definition().build(repo=repo)

    assert dryml.config.build_repo is None

    assert obj.definition() == obj2.definition()
    assert obj.A is obj.B
    assert obj2.A is obj2.B
    assert obj.A is obj2.A
    assert obj.B is obj2.B
    assert obj.A is obj2.B


@pytest.mark.usefixtures("create_temp_dir")
def test_object_save_restore_with_repo_2(create_temp_dir):
    """
    We test save and restore of nested objects with a repo
    """
    import objects

    repo = dryml.Repo(create_temp_dir, create=True)

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    # Add and save object in repo
    repo.add_object(data_obj1)
    repo.save()

    # Enclose them in another object
    obj = objects.TestClassC(data_obj1, B=data_obj1)

    # Save the enclosing object.
    repo.add_object(obj)
    repo.save()

    # There should now be two objects stored.
    assert len(repo) == 2

    # Get top object definition
    obj_def = obj.definition()

    repo2 = dryml.Repo(create_temp_dir)

    obj2 = obj_def.build(repo=repo2)

    assert obj_def == obj2.definition()
    assert obj2.A is obj2.B
    assert obj.A.C == obj2.A.C
    assert obj.B.C == obj2.B.C
    assert obj.A.data == obj2.A.data
    assert obj.B.data == obj2.B.data


@pytest.mark.usefixtures("create_temp_dir")
def test_object_save_restore_with_repo_3(create_temp_dir):
    """
    We test save and restore of nested objects with a repo
    """
    import objects

    # Create workshop
    repo = dryml.Repo(directory=create_temp_dir)

    obj_a = objects.TestNest(10)
    repo.add_object(obj_a)

    def build_def(repo):
        # Create the data containing objects
        obj_a = dryml.utils.head(repo.get(
            selector=dryml.ObjectDef(objects.TestNest, 10)))

        mdl_def = dryml.ObjectDef(
            objects.TestNest2,
            A=10)

        mdl_def = dryml.ObjectDef(
            objects.TestNest3,
            model=mdl_def)

        mdl_def = dryml.ObjectDef(
            objects.TestNest3,
            obj_a,
            mdl_def)

        return mdl_def

    model_def = build_def(repo)

    @dryml.compute_context(ctx_context_reqs={'default': {}})
    def test_method(model_def, location):
        # Create repo
        repo = dryml.Repo(directory=location)

        # Build the object
        model_obj = model_def.build(repo=repo)

        # Save all objects
        repo.save(model_obj)

    test_method(model_def, create_temp_dir)

    repo.load_objects_from_directory()

    repo.get(model_def, sel_kwargs={'verbosity': 2})



import pytest
import os
from dryml import List, Tuple, Dict, Repo, load_object
import objects


def test_dry_list_1():
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")
    the_list = List(obj1, obj2)

    new_list = the_list.definition().build()

    assert new_list[0].definition() == obj1.definition()
    assert new_list[1].definition() == obj2.definition()
    assert new_list is not the_list
    assert new_list[0] is not obj1
    assert new_list[1] is not obj2
    assert the_list.definition().dry_mut


def test_dry_list_2():
    # Create a repo
    repo = Repo()

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")

    # Add them to the repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create a list object
    the_list = List(obj1, obj2)

    # Create a new list object from a definition, but use a repo
    new_list = the_list.definition().build(repo=repo)

    # The list object should've extracted the objects themselves
    # From the repo instead of constructing them
    assert the_list[0] is new_list[0]
    assert the_list[1] is new_list[1]

    # BUT we should have a 'new' list object.
    assert the_list is not new_list


def test_dry_list_3():
    # Create a repo
    repo = Repo()

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")

    # Add them to the repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create a list object
    the_list = List(obj1, obj2)

    # Add that to the repo
    repo.add_object(the_list)

    # Create a new list object from a definition, but use a repo
    new_list = the_list.definition().build(repo=repo)

    # The list object should've extracted the objects themselves
    # From the repo instead of constructing them
    assert the_list[0] is new_list[0]
    assert the_list[1] is new_list[1]

    # BUT we should not have a 'new' list object.
    assert the_list is new_list


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_list_4(create_temp_dir):
    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj1.save_self(os.path.join(create_temp_dir, 'obj1.dry'))
    obj2 = objects.HelloStr(msg="a test")
    obj2.save_self(os.path.join(create_temp_dir, 'obj2.dry'))

    # Create a list object
    the_list = List(obj1, obj2)

    # Save the list object
    the_list.save_self(os.path.join(create_temp_dir, 'obj_list.dry'))

    # Load list from disk without a repo
    new_list = load_object(os.path.join(create_temp_dir, 'obj_list.dry'))

    assert len(the_list) == len(new_list)
    assert the_list[0].definition() == new_list[0].definition()
    assert the_list[1].definition() == new_list[1].definition()
    assert the_list[0] is not new_list[0]
    assert the_list[1] is not new_list[1]


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_list_5(create_temp_dir):
    # Create repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")

    # Add objects to the repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create a list object
    the_list = List(obj1, obj2)

    # Add the list to the repo
    repo.add_object(the_list, filepath="obj_list.dry")

    # Save objects
    repo.save()

    new_list = load_object(os.path.join(create_temp_dir, 'obj_list.dry'),
                           repo=repo)

    assert the_list is new_list


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_list_6(create_temp_dir):
    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    list_1 = List(obj1)
    list_2 = List(obj2, list_1)

    list_2.save_self(os.path.join(create_temp_dir, 'list.dry'))

    loaded_list = load_object(os.path.join(create_temp_dir, 'list.dry'))

    assert list_2.definition() == loaded_list.definition()


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_list_7(create_temp_dir):
    # Create Repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    # Add objects to Repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create lists
    list_1 = List(obj1)

    # Add list to repo
    repo.add_object(list_1)

    # Create second list, and save it
    list_2 = List(obj2, list_1)
    list_2.save_self(os.path.join(create_temp_dir, 'list.dry'))

    loaded_list = load_object(os.path.join(create_temp_dir, 'list.dry'),
                              repo=repo)

    assert list_2.definition() == loaded_list.definition()
    assert list_2 is not loaded_list
    assert list_2[0] is loaded_list[0]
    assert list_2[1] is loaded_list[1]


def test_dry_tuple_1():
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")
    the_tuple = Tuple(obj1, obj2)

    new_tuple = the_tuple.definition().build()
    assert new_tuple[0].definition() == obj1.definition()
    assert new_tuple[1].definition() == obj2.definition()
    assert new_tuple is not the_tuple
    assert new_tuple[0] is not obj1
    assert new_tuple[1] is not obj2
    assert not new_tuple.definition().dry_mut


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_tuple_2(create_temp_dir):
    # Create Repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    # Add objects to Repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create lists
    tuple_1 = Tuple(obj1)

    # Add list to repo
    repo.add_object(tuple_1)

    # Create second list, and save it
    tuple_2 = Tuple(obj2, tuple_1)
    tuple_2.save_self(os.path.join(create_temp_dir, 'tuple.dry'))

    new_tuple = tuple_2.definition().build(repo=repo)

    assert tuple_2.definition() == new_tuple.definition()
    assert tuple_2 is not new_tuple
    assert tuple_2[0] is new_tuple[0]
    assert tuple_2[1] is new_tuple[1]


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_tuple_3(create_temp_dir):
    # Create Repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    # Add objects to Repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create lists
    tuple_1 = Tuple(obj1)

    # Add list to repo
    repo.add_object(tuple_1)

    # Create second list, and save it
    tuple_2 = Tuple(obj2, tuple_1)
    tuple_2.save_self(os.path.join(create_temp_dir, 'tuple.dry'))

    loaded_tuple = load_object(os.path.join(create_temp_dir, 'tuple.dry'),
                               repo=repo)

    assert tuple_2.definition() == loaded_tuple.definition()
    assert tuple_2 is not loaded_tuple
    assert tuple_2[0] is loaded_tuple[0]
    assert tuple_2[1] is loaded_tuple[1]


def test_dry_tuple_4():
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.TestNest2(A=obj1)
    obj3 = objects.HelloInt(msg=10)
    obj4 = objects.TestNest2(A=obj3)
    tuple_1 = Tuple(obj2, obj4)
    tuple_def = tuple_1.definition()
    tuple_2 = tuple_def.build()
    assert tuple_1[0].A.int_msg == tuple_2[0].A.int_msg
    assert tuple_1[1].A.int_msg == tuple_2[1].A.int_msg


def test_dry_dict_1():
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg="a test")
    the_dict = Dict({'a': obj1, 2: obj2})

    # Assert storage works properly
    assert the_dict['a'] is obj1
    assert the_dict[2] is obj2

    # Build new dict copy
    new_dict = the_dict.definition().build()
    assert new_dict['a'].definition() == obj1.definition()
    assert new_dict[2].definition() == obj2.definition()
    assert new_dict is not the_dict
    assert new_dict['a'] is not obj1
    assert new_dict[2] is not obj2
    assert new_dict.definition().dry_mut


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_dict_2(create_temp_dir):
    # Create Repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    # Add objects to Repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create dicts
    dict_1 = Dict({'a': obj1})

    # Add list to repo
    repo.add_object(dict_1)

    # Create second list, and save it
    dict_2 = Dict({'b': obj2, 1.0: dict_1})
    dict_2.save_self(os.path.join(create_temp_dir, 'dict.dry'))

    new_dict = dict_2.definition().build(repo=repo)

    assert dict_2.definition() == new_dict.definition()
    assert dict_2 is not new_dict
    assert dict_2['b'] is new_dict['b']
    assert dict_2[1.0] is new_dict[1.0]


@pytest.mark.usefixtures("create_temp_dir")
def test_dry_dict_3(create_temp_dir):
    # Create Repo
    repo = Repo(create_temp_dir)

    # Create objects
    obj1 = objects.HelloInt(msg=5)
    obj2 = objects.HelloStr(msg='a test')

    # Add objects to Repo
    repo.add_object(obj1)
    repo.add_object(obj2)

    # Create lists
    dict_1 = Dict({'a': obj1})

    # Add list to repo
    repo.add_object(dict_1)

    # Create second list, and save it
    dict_2 = Dict({'b': obj2, 2.0: dict_1})
    dict_2.save_self(os.path.join(create_temp_dir, 'dict.dry'))

    loaded_dict = load_object(os.path.join(create_temp_dir, 'dict.dry'),
                              repo=repo)

    assert dict_2.definition() == loaded_dict.definition()
    assert dict_2 is not loaded_dict
    assert dict_2['b'] is loaded_dict['b']
    assert dict_2[2.0] is loaded_dict[2.0]



from dryml.context.context_tracker import context, set_context, \
    contexts, ComputeContext, get_context_class, context_check, \
    ResourcesUnavailableError, WrongContextError, get_context_manager, \
    NoContextError, ContextAlreadyActiveError, get_context_requirements, \
    ResourcePool, ResourceRequest, ResourceAllocation, \
    InsufficientResourcesError, ContextManager, ContextContainer, \
    ContextIncompatibilityError, combine_requests, combine_reqs
from dryml.context.process import Process, compute_context, compute, \
    cls_method_compute, tune_compute_context


__all__ = [
    context,
    set_context,
    get_context_class,
    get_context_manager,
    get_context_requirements,
    context_check,
    contexts,
    combine_requests,
    combine_reqs,
    ComputeContext,
    ContextManager,
    ContextContainer,
    ResourcesUnavailableError,
    WrongContextError,
    NoContextError,
    ContextAlreadyActiveError,
    ContextIncompatibilityError,
    Process,
    compute_context,
    tune_compute_context,
    compute,
    cls_method_compute,
    ResourcePool,
    ResourceRequest,
    ResourceAllocation,
    InsufficientResourcesError,
]



# File to define saving/loading convenience functions

from __future__ import annotations

import os
import dill
import pickle
import io
import zipfile
import uuid
import re
import numpy as np
import time

from typing import IO, Union, Optional, Type, Callable
from dryml.config import ObjectDef, Meta, MissingIdError, MissingMetadataError
from dryml.utils import get_current_cls, pickler, static_var, \
    is_supported_scalar_type, is_supported_listlike, is_supported_dictlike, \
    map_dictlike, map_listlike, get_class_from_str, get_class_str
from dryml.context.context_tracker import combine_requests, context, \
    NoContextError
from dryml.file_intermediary import FileIntermediary
from dryml.save_cache import SaveCache


FileType = Union[str, IO[bytes]]


def file_resolve(file: str, exact_path: bool = False) -> str:
    if os.path.splitext(file)[1] == '' and not exact_path:
        file = f"{file}.dry"
    return file


class ObjectFile(object):
    contained_dry_file_re = re.compile(r"^dry_objects/([a-f0-9-]*)\.dry$")

    # Supports 'save cached' file writing.
    def __init__(self, file: FileType, exact_path: bool = False,
                 mode: str = 'r', must_exist: bool = True,
                 save_cache=None, save_caching=True):

        if type(file) is zipfile.ZipFile:
            raise TypeError(
                "Passing zipfiles directly is currently not supported.")

        self.mode = mode

        # If file is a string, resolve it to a filepath, and save this filepath
        if type(file) is str:
            filepath = file
            filepath = file_resolve(filepath, exact_path=exact_path)
            if must_exist and not os.path.exists(filepath):
                raise ValueError(f"File {filepath} doesn't exist!")
            self.filepath = filepath

        if self.mode == 'w':
            self._z_file = None
            if hasattr(self, 'filepath'):
                self.binary_file = open(self.filepath, 'wb')
            else:
                self.binary_file = file
        elif self.mode == 'r':
            if hasattr(self, 'filepath'):
                self.binary_file = open(self.filepath, 'rb')
                self._z_file = zipfile.ZipFile(self.binary_file, mode=mode)
            else:
                if type(file) is zipfile.ZipFile:
                    self._z_file = file
                    if self._z_file.mode != mode:
                        raise ValueError(
                            "Utilized Zipfile doesn't match requested mode!")
                else:
                    self._z_file = zipfile.ZipFile(file, mode=mode)
        else:
            raise ValueError(f"Unsupported mode: {mode}")

    @property
    def z_file(self):
        # We need this property to create the zip file on demand
        # since sometimes We might write directly to the binary file.
        if self.mode == 'w':
            # Check if we've already opened a zip file.
            if self._z_file is not None:
                return self._z_file

            self.int_file = FileIntermediary()
            self.close_int_file = True

            self._z_file = zipfile.ZipFile(self.int_file, mode=self.mode)
            return self._z_file
        elif self.mode == 'r':
            return self._z_file

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def int_file_detach(self):
        if not hasattr(self, 'int_file'):
            raise RuntimeError(
                "ObjectFile doesnt' have an intermediate file to detach.")
        self.close_int_file = False
        return self.int_file

    def close(self):
        # Close the zipfile.
        if self._z_file is not None:
            # Avoid accessing z_file property if it doesn't already exist.
            # Accessing the property triggers the int_file to be created,
            # And an empty file to be written out.
            self.z_file.close()

        # If we have a binary open, sync any intermediary and close it.
        if hasattr(self, 'binary_file'):
            if self.mode == 'w':
                if hasattr(self, 'int_file'):
                    # If there's an intermediary file open, write it to the
                    # binary file.
                    self.int_file.write_to_file(self.binary_file)
            if hasattr(self, 'filepath'):
                # We opened the binary file. we should close it.
                self.binary_file.close()

        # Close the intermediary if needed.
        if hasattr(self, 'int_file'):
            if self.close_int_file:
                self.int_file.close()

    # def update_file(self, obj: Object):
    #     self.cache_object_data_obj(obj)

    def save_meta_data(self):
        # Meta_data
        meta_data = {
            'version': 1
        }

        meta_dump = pickler(meta_data)
        with self.z_file.open('meta_data.pkl', mode='w') as f:
            f.write(meta_dump)

    def load_meta_data(self):
        try:
            with self.z_file.open('meta_data.pkl', 'r') as meta_file:
                meta_data = pickle.loads(meta_file.read())
        except KeyError as e:
            print(f"Issue getting MetaData! Zipfile contains the "
                  f"following files: {self.z_file.namelist()}")
            print(f"Zipfile: {self.z_file}")
            raise e
        return meta_data

    def save_class_def_v1(self, obj_def: ObjectDef, update: bool = False):
        # We need to pickle the class definition.
        # By default, error out if class has changed. Check this.
        mod_cls = get_current_cls(obj_def.cls)
        if obj_def.cls != mod_cls and not update:
            raise ValueError("Can't save class definition! It's been changed!")
        try:
            cls_def = dill.dumps(mod_cls)
            with self.z_file.open('cls_def.dill', mode='w') as f:
                f.write(cls_def)
        except TypeError as e:
            if '_abc_data' in str(e):
                # Fallback to pickling the class name.
                # In some cases, the class's _abc_impl
                # attribute seems wrongly constructed?
                # https://github.com/uqfoundation/dill/issues/332
                cls_str = get_class_str(mod_cls)
                with self.z_file.open('cls_str.txt', mode='w') as f:
                    f.write(cls_str.encode('utf-8'))
            else:
                raise e

    def load_class_def_v1(self, update: bool = True, reload: bool = False):
        "Helper function for loading a version 1 class definition"
        namelist = self.z_file.namelist()
        # Get class definition
        if 'cls_def.dill' in namelist:
            with self.z_file.open('cls_def.dill') as cls_def_file:
                if update:
                    # Get original model definition
                    cls_def_init = dill.loads(cls_def_file.read())
                    try:
                        cls_def = get_current_cls(cls_def_init, reload=reload)
                    except Exception as e:
                        raise RuntimeError(
                            f"Failed to update module class {e}")
                else:
                    cls_def = dill.loads(cls_def_file.read())
            return cls_def
        elif 'cls_str.txt' in namelist:
            with self.z_file.open('cls_str.txt') as cls_str_file:
                cls_str = cls_str_file.read().decode('utf-8')
                cls_def = get_class_from_str(cls_str, reload=reload)
            return cls_def
        else:
            raise RuntimeError("No stored class data!")

    def save_definition_v1(self, obj_def: ObjectDef, update: bool = False):
        "Save object def"
        # Save obj def
        self.save_class_def_v1(obj_def, update=update)

        # Save args from object def
        with self.z_file.open('dry_args.pkl', mode='w') as args_file:
            args_file.write(pickler(obj_def.args))

        # Save kwargs from object def
        with self.z_file.open('dry_kwargs.pkl', mode='w') as kwargs_file:
            kwargs_file.write(pickler(obj_def.kwargs))

        # Save mutability from object def
        with self.z_file.open('dry_mut.pkl', mode='w') as mut_file:
            mut_file.write(pickler(obj_def.dry_mut))

    def load_definition_v1(self, update: bool = True, reload: bool = False):
        "Load object def"

        # Load obj def
        cls = self.load_class_def_v1(update=update, reload=reload)

        # Load args
        with self.z_file.open('dry_args.pkl', mode='r') as args_file:
            args = pickle.loads(args_file.read())

        # Load kwargs
        with self.z_file.open('dry_kwargs.pkl', mode='r') as kwargs_file:
            kwargs = pickle.loads(kwargs_file.read())

        # Load mutability
        with self.z_file.open('dry_mut.pkl', mode='r') as mut_file:
            mut = pickle.loads(mut_file.read())

        return ObjectDef(cls, *args, dry_mut=mut, **kwargs)

    def definition(self, update: bool = True, reload: bool = False):
        meta_data = self.load_meta_data()
        if meta_data['version'] == 1:
            return self.load_definition_v1(update=update, reload=reload)
        else:
            raise RuntimeError(
                f"File version {meta_data['version']} not supported!")

    def load_object_v1(self, update: bool = True,
                       reload: bool = False,
                       as_cls: Optional[Type] = None) -> Object:
        # Load object
        obj_def = self.load_definition_v1(update=update, reload=reload)
        if as_cls is not None:
            obj_def.cls = as_cls

        # Create object
        obj = obj_def.build(load_zip=self.z_file)

        # Load object content
        if not obj.load_object(self.z_file):
            raise RuntimeError("Error loading object!")

        # Build object instance
        return obj

    def load_object_content(self, obj: Object) -> bool:
        file_def = self.definition()
        obj_def = obj.definition()
        if file_def != obj_def or file_def.dry_id != obj_def.dry_id:
            raise ValueError(
                f"File {self.z_file} doesn't store data for object "
                f"{obj.dry_id} at the top level.")

        if not obj.load_object(self.z_file):
            return False
        return True

    def load_object(self, update: bool = False,
                    reload: bool = False,
                    as_cls: Optional[Type] = None) -> Object:
        meta_data = self.load_meta_data()
        version = meta_data['version']
        if version == 1:
            return self.load_object_v1(
                update=update, reload=reload, as_cls=as_cls)
        else:
            raise RuntimeError(f"DRY version {version} unknown")

    def save_object_v1(self, obj: Object, update: bool = False,
                       as_cls: Optional[Type] = None,
                       save_cache=None) -> bool:

        # First, check the save cache.
        if save_cache is not None:
            # WARNING, the id of an object is only unique for its lifetime.
            # We must assume objects aren't going out of scope
            # or being deleted for this to work.
            if id(obj) in save_cache.obj_cache:
                # We found the object, write the cached file to
                # The passed binary.
                saved_int_file = save_cache.obj_cache[id(obj)]
                saved_int_file.write_to_file(self.binary_file)
                return True

        # Save subordinate objects.
        for sub_obj in obj.__dry_obj_container_list__:
            # Open a file inside the zip to contain the new object.
            obj_id = sub_obj.dry_id
            save_path = f'dry_objects/{obj_id}.dry'
            if save_path not in self.z_file.namelist():
                with self.z_file.open(save_path, 'w') as f:
                    if not sub_obj.save_self(f, save_cache=save_cache):
                        return False

        # Save meta data
        self.save_meta_data()

        # Save config v1
        obj_def = obj.definition()
        if as_cls is not None:
            obj_def.cls = as_cls

        self.save_definition_v1(obj_def, update=update)

        # Save object content
        ret_val = obj.save_object(self.z_file, save_cache=save_cache)

        if save_cache is not None:
            save_cache.obj_cache[id(obj)] = self.int_file_detach()

        return ret_val

    def contained_object_ids(self):
        """
        enumerates the ids of contained subordinate objects
        """

        return list(map(
            lambda m: m.groups(1)[0],
            filter(lambda m: m is not None,
                   map(lambda n: ObjectFile.contained_dry_file_re.match(n),
                       self.z_file.namelist()))))

    def get_contained_object_file(self, dry_id):
        return self.z_file.open(f"dry_objects/{dry_id}.dry")


@static_var('load_repo', None)
def load_object(file: FileType, update: bool = False,
                exact_path: bool = False,
                reload: bool = False,
                as_cls: Optional[Type] = None,
                repo=None) -> Object:
    """
    A method for loading an object from disk.
    """
    reset_repo = False
    load_obj = True

    # Handle repo management variables
    if repo is not None:
        if load_object.load_repo is not None:
            raise RuntimeError(
                "different repos not currently supported")
        else:
            # Set the call_repo
            load_object.load_repo = repo
            reset_repo = True

    # We now need the object definition
    with ObjectFile(file, exact_path=exact_path) as dry_file:
        obj_def = dry_file.definition()
        # Check whether a repo was given in a prior call
        if load_object.load_repo is not None:
            try:
                # Load the object from the repo
                obj = load_object.load_repo.get_obj(obj_def)
                if obj.definition() != obj_def:
                    raise RuntimeError("Found issue!")
                load_obj = False
            except Exception:
                pass

        if load_obj:
            obj = dry_file.load_object(update=update,
                                       reload=reload,
                                       as_cls=as_cls)
            cls_str_compare = False
            if as_cls is not None or reload:
                if as_cls is not None:
                    cls = as_cls
                elif reload:
                    cls = get_class_from_str(get_class_str(obj_def.cls))
                cls_str_compare = True
                new_def = ObjectDef(
                    cls,
                    *obj_def.args,
                    dry_mut=obj_def.dry_mut,
                    **obj_def.kwargs)
            else:
                new_def = obj_def
            if not obj.definition().equal(
                    new_def, cls_str_compare=cls_str_compare):
                raise RuntimeError(
                    f"Loaded object doesn't have expected definition! "
                    f"expected {new_def} got: {obj.definition()}")

    # Reset the repo for this function
    if reset_repo:
        load_object.load_repo = None

    return obj


@static_var('load_repo', None)
def load_object_content(
        obj: Object,
        file: FileType) -> bool:
    """
    A method for loading an object from disk.
    """

    # We now need the object definition
    with ObjectFile(file) as dry_file:
        file_def = dry_file.definition()
        obj_def = obj.definition()
        if file_def != obj_def or file_def.dry_id != obj_def.dry_id:
            raise ValueError(
                f"File {file} doesn't store data for object {obj.dry_id} "
                "at the top level.")

        # Load contained objects
        file_contained_obj_ids = dry_file.contained_object_ids()
        for sub_obj in obj.__dry_obj_container_list__:
            if sub_obj.dry_id not in file_contained_obj_ids:
                raise ValueError(
                    f"File {file} doesn't contain subordinate object data for "
                    f"subordinate object {sub_obj.dry_id}! file contains: "
                    f"{file_contained_obj_ids} namelist: "
                    f"{dry_file.z_file.namelist()}")
            sub_obj_f = dry_file.get_contained_object_file(sub_obj.dry_id)
            if not load_object_content(sub_obj, sub_obj_f):
                print(f"Error loading subordinate object {sub_obj.dry_id}")
                return False

        # Load content of this object
        if not dry_file.load_object_content(obj):
            print(f"Error loading self: {obj.dry_id}")
            return False

    return True


def save_object(obj: Object, file: FileType, version: int = 1,
                exact_path: bool = False, update: bool = False,
                as_cls: Optional[Type] = None,
                save_cache=None) -> bool:
    # Initialize a save cache by default.
    close_save_cache = False
    if save_cache is None:
        close_save_cache = True
        save_cache = SaveCache()
    with ObjectFile(file, exact_path=exact_path, mode='w',
                    must_exist=False) as dry_file:
        if version == 1:
            ret_val = dry_file.save_object_v1(
                obj, update=update, as_cls=as_cls, save_cache=save_cache)
        else:
            raise ValueError(f"File version {version} unknown. Can't save!")

    # Close save caches.
    if save_cache is not None and close_save_cache:
        del save_cache

    return ret_val


def change_object_cls(obj: Object, cls: Type, update: bool = False,
                      reload: bool = False) -> Object:
    buffer = io.BytesIO()
    if not save_object(obj, buffer):
        raise RuntimeError("Error saving object!")
    return load_object(buffer, update=update, reload=reload,
                       as_cls=cls)


def obj_to_def(val):
    if is_supported_scalar_type(val):
        return val
    if isinstance(val, Object):
        definition = val.definition()
        return definition
    elif is_supported_listlike(val):
        return map_listlike(obj_to_def, val)
    elif is_supported_dictlike(val):
        return map_dictlike(obj_to_def, val)
    else:
        raise TypeError(
            f"Unsupported value {val} of type "
            f"{type(val)} encountered!")


# Define a base  Object
class Object(metaclass=Meta):
    # Only ever set for this class.
    __dry_meta_base__ = True

    # Define the dry_id
    def __init__(self, *args, dry_id=None, dry_metadata=None, **kwargs):
        # Dry ID
        if dry_id is None:
            self.dry_kwargs['dry_id'] = str(uuid.uuid4())

        # Handle Metadata
        if dry_metadata is None:
            # Create a default empty dry_metadata
            dry_metadata = {}
        if 'description' not in dry_metadata:
            dry_metadata['description'] = ""
        if 'creation_time' not in dry_metadata:
            dry_metadata['creation_time'] = time.time()
        self.dry_kwargs['dry_metadata'] = dry_metadata

        self._definition = None

    def definition(self):
        # Return if we have a definition created already
        if self._definition is not None:
            return self._definition

        # recursively build the definitions
        new_args = obj_to_def(self.dry_args)
        new_kwargs = obj_to_def(self.dry_kwargs)

        self._definition = ObjectDef(
            type(self),
            *new_args,
            **new_kwargs)

        return self._definition

    def save_self(self, file: FileType, version: int = 1, **kwargs) -> bool:
        return save_object(self, file, version=version, **kwargs)

    def __str__(self):
        return str(self.definition())

    def __repr__(self):
        return str(self.definition())

    @property
    def dry_id(self):
        if 'dry_id' not in self.dry_kwargs:
            raise MissingIdError()
        return self.dry_kwargs['dry_id']

    @property
    def dry_metadata(self):
        if 'dry_metadata' not in self.dry_kwargs:
            raise MissingMetadataError()
        return self.dry_kwargs['dry_metadata']

    def __hash__(self):
        return hash(self.dry_id)

    def dry_context_requirements(self):
        context_reqs = {self.__dry_compute_context__: [{}]}
        for obj in self.__dry_obj_container_list__:
            obj_reqs = obj.dry_context_requirements()
            for ctx_name in obj_reqs:
                if ctx_name in context_reqs:
                    context_reqs[ctx_name].append(obj_reqs[ctx_name])
                else:
                    context_reqs[ctx_name] = [obj_reqs[ctx_name]]

        for ctx_name in context_reqs:
            context_reqs[ctx_name] = combine_requests(context_reqs[ctx_name])

        return context_reqs

    def compute_activate(self):
        ctx_mgr = context()
        if ctx_mgr is None:
            raise NoContextError()

        # Activate contained objects first.
        # We can activate subordinate objects and depend
        # on them being loaded.
        if hasattr(self, '__dry_obj_container_list__'):
            for obj in self.__dry_obj_container_list__:
                obj.compute_activate()

        # Prepare this object's compute
        if not ctx_mgr.contains_activated_object(self):
            self.compute_prepare()

            # Load this object's compute if needed
            self.load_compute()

            # Add object to manager tracker.
            ctx_mgr.add_activated_object(self)

    @staticmethod
    def graph_label(obj, report_class=True):
        if report_class:
            return f"{obj.dry_id} ({type(obj).__name__})"
        else:
            return obj.dry_id

    def _dry_obj_graph(self, report_class=True):
        root = {}
        if hasattr(self, '__dry_obj_container_list__') and \
           len(self.__dry_obj_container_list__) > 0:
            for obj in self.__dry_obj_container_list__:
                label = Object.graph_label(
                    obj, report_class=report_class)
                root[label] = obj._dry_obj_graph()
        return root

    def dry_obj_graph(self, report_class=True):
        from asciitree import LeftAligned
        from asciitree.drawing import BoxStyle, BOX_LIGHT
        label = Object.graph_label(
            self, report_class=report_class)
        tree = {label: self._dry_obj_graph()}
        tr = LeftAligned(
            draw=BoxStyle(
                gfx=BOX_LIGHT,
                horiz_len=1
            )
        )
        print(tr(tree))


class ObjectFactory(object):
    def __init__(self, obj_def: ObjectDef, callbacks=[]):
        if 'dry_id' in obj_def:
            raise ValueError(
                "An Object factory can't use a definition with a dry_id")
        self.obj_def = obj_def
        self.callbacks = callbacks

    def add_callback(self, callback):
        self.callbacks.append(callback)

    def __call__(self, repo=None):
        obj = self.obj_def.build(repo=repo)
        for callback in self.callbacks:
            # Call each callback
            callback(obj)
        return obj


class ObjectWrapper(Object):
    """
    An object wrapper for simple python objects
    """
    def __init__(self, cls: Type, obj_args=None, obj_kwargs=None):
        if obj_args is None:
            obj_args = []
            self.dry_kwargs['obj_args'] = obj_args
        if obj_kwargs is None:
            obj_kwargs = {}
            self.dry_kwargs['obj_kwargs'] = obj_kwargs

        self.obj = cls(*obj_args, **obj_kwargs)


class CallableWrapper(Object):
    """
    A wrapper for a callable object to cement some arguments
    """

    def __init__(
            self, obj: ObjectWrapper, obj_args=None, obj_kwargs=None,
            call_args=None, call_kwargs=None):
        if obj_args is None:
            obj_args = []
            self.dry_kwargs['obj_args'] = obj_args
        if obj_kwargs is None:
            obj_kwargs = {}
            self.dry_kwargs['obj_kwargs'] = obj_kwargs
        if call_args is None:
            call_args = []
            self.dry_kwargs['call_args'] = call_args
        self.call_args = call_args
        if call_kwargs is None:
            call_kwargs = {}
            self.dry_kwargs['call_kwargs'] = call_kwargs
        self.call_kwargs = call_kwargs
        self.obj = obj

    def __call__(self, *args, **kwargs):
        return self.obj.obj(
            *(self.call_args+args),
            **{**self.call_kwargs, **kwargs})


def get_contained_objects(obj: Object) -> [Object]:
    contained_objs = set()
    for contained_obj in obj.__dry_obj_container_list__:
        sub_set = get_contained_objects(contained_obj)
        # Add contained objects contained objects.
        contained_objs.update(sub_set)
        # Add the contained object.
        contained_objs.add(contained_obj)
    return contained_objs


class DryObjectPlaceholder(object):
    def __init__(self, ID, obj_def):
        self.ID = ID
        self.obj_def = obj_def


class DryObjectPlaceholderData(object):
    def __init__(self, ID: int, data):
        self.ID = ID
        self.data = data


def generate_unique_id():
    ii32 = np.iinfo(np.int32)
    return np.random.randint(ii32.min, ii32.max)


def create_placeholder(obj: Object) -> (DryObjectPlaceholder,
                                        DryObjectPlaceholderData):
    ID = generate_unique_id()
    obj_def = obj.definition()
    ph_def = DryObjectPlaceholder(ID, obj_def)
    temp_buf = io.BytesIO()
    if not obj.save_self(temp_buf):
        raise RuntimeError(
            "Couldn't save object {obj.definition()} to temp buffer!")

    temp_buf.seek(0)
    ph_data = DryObjectPlaceholderData(ID, temp_buf.read())
    return (ph_def, ph_data)


def rebuild_object(ph_def, ph_data, verbose=False) -> Object:
    if verbose:
        print(f"Rebuilding object {ph_def} with data {ph_data}")
    obj = ph_def.obj_def.build()
    buf = io.BytesIO(ph_data.data)
    if not load_object_content(obj, buf):
        raise RuntimeError(f"Failed to rebuild object {ph_def.obj_def}")
    return obj


def prep_args_kwargs(args, kwargs):
    obj_ph_data_map = {}
    obj_ph_map = {}
    new_args = []
    for i in range(len(args)):
        arg = args[i]
        assigned = False
        if issubclass(type(arg), Object):
            if id(arg) in obj_ph_data_map:
                new_args.append(obj_ph_map[id(arg)])
            else:
                ph = create_placeholder(arg)
                new_args.append(ph[0])
                obj_ph_map[id(arg)] = ph[0]
                obj_ph_data_map[id(arg)] = ph[1]
            assigned = True
        if not assigned:
            new_args.append(args[i])

    for key in kwargs:
        arg = kwargs[key]
        if issubclass(type(arg), Object):
            if id(arg) in obj_ph_data_map:
                args[i] = obj_ph_map[id(arg)]
            else:
                ph = create_placeholder(arg)
                args[i] = ph[0]
                obj_ph_map[id(arg)] = ph[0]
                obj_ph_data_map[id(arg)] = ph[1]

    return (new_args, kwargs), list(obj_ph_data_map.values())


def reconstruct_args_kwargs(args, kwargs, ph_data, verbose=False):
    ph_dict = {}
    constructed_objs = {}
    for ph in ph_data:
        ph_dict[ph.ID] = ph

    for i in range(len(args)):
        arg = args[i]
        if type(arg) is DryObjectPlaceholder:
            if arg.ID in constructed_objs:
                args[i] = constructed_objs[arg.ID]
            else:
                obj = rebuild_object(arg, ph_dict[arg.ID], verbose=verbose)
                constructed_objs[arg.ID] = obj
                args[i] = obj

    for key in kwargs:
        arg = kwargs[key]
        if type(arg) is DryObjectPlaceholder:
            if arg.ID in constructed_objs:
                kwargs[key] = constructed_objs[arg.ID]
            else:
                obj = rebuild_object(arg, ph_dict[arg.ID], verbose=verbose)
                constructed_objs[arg.ID] = obj
                kwargs[key] = obj


class ObjectNode(object):
    """
    Helper class to Build tree of objects
    """

    def __init__(self, obj, children):
        # All nodes have a list of children
        self.children = children
        self.obj = obj

    def apply_func(self, func: Callable):
        if self.obj is not None:
            func(self.obj)


def apply_df_imp(func: Callable, node: ObjectNode, seen_set: set[ObjectNode]):
    # seen set will be empty when called on the root node.
    for child_node in node.children:
        if child_node not in seen_set:
            # Apply function to child node.
            apply_df_imp(func, child_node, seen_set)
    # Apply to this node
    node.apply_func(func)
    # Set this node as seen.
    seen_set.add(node)


def apply_bf_imp(
        func: Callable,
        visit_list: list[ObjectNode], seen_set: set[ObjectNode]):
    # the visit list should be populated when called from the root node.
    while len(visit_list) > 0:
        # Pop a node from the front of the list.
        node = visit_list.pop(0)
        # Apply function first to this node.
        node.apply(func)
        # Add this node to the seen set.
        seen_set.add(node)
        # add nodes children that haven't already been seen to the visit list.
        for child_node in node.children:
            if child_node not in seen_set:
                visit_list.append(child_node)


class ObjectTree(ObjectNode):
    """
    Root node object.
    """

    def __init__(self, children):
        super().__init__(None, children)

    def apply_func(self, func: Callable):
        # Don't do anything on the root node.
        pass

    def apply_df(self, func: Callable):
        apply_df_imp(func, self, set())

    def apply_bf(self, func: Callable):
        apply_bf_imp(func, self, self.children, set())


def build_obj_tree_imp(obj: Object, node_dict):
    """
    Returns ObjectNode of passed object.
    """

    if id(obj) in node_dict:
        return node_dict[id(obj)]

    # Get unique children list
    unique_children_map = {}
    for child_obj in obj.__dry_obj_container_list__:
        if id(child_obj) not in unique_children_map:
            unique_children_map[id(child_obj)] = child_obj

    # Build child node list
    child_nodes = []
    for child_id in unique_children_map:
        child_obj = unique_children_map[child_id]
        child_nodes.append(build_obj_tree_imp(child_obj, node_dict))

    # Create node, and record.
    res_node = ObjectNode(obj, child_nodes)
    node_dict[id(obj)] = res_node
    return res_node


def build_obj_tree(obj_or_list: Union[Object, list[Object]]):
    """
    Construct tree
    """
    # We construct depth first.
    node_dict = {}

    if type(obj_or_list) is Object:
        obj_list = [obj_or_list]
    else:
        obj_list = obj_or_list

    # Get unique objects
    obj_dict = {}
    for obj in obj_list:
        if id(obj) not in obj_dict:
            obj_dict[id(obj)] = obj

    obj_list = obj_dict.values()

    root_children = []
    for obj in obj_list:
        root_children.append(build_obj_tree_imp(obj, node_dict))

    # Returned finished tree
    return ObjectTree(root_children)



import os
import pathlib
import dryml
from dryml.models import TrainSpec
from dryml.models import Trainable
from dryml.data import Dataset
from dryml import Repo
import pickle
import uuid


class Tune2ObjectSaver(object):
    def __init__(
            self, model: Trainable = None,
            train_state: TrainSpec = None,
            repo: Repo = None,
            test_ds: Dataset = None,
            ctx_reqs=None,
            tmp_checkpoint_dir='/tmp',
            metrics={}):
        if model is None:
            raise ValueError("Must pass a model.")
        self.model = model

        if train_state is None:
            raise ValueError("Must pass a train state.")
        self.train_state = train_state

        if repo is None:
            raise ValueError("Must pass a repo.")
        self.repo = repo

        if test_ds is None:
            raise ValueError("Must pass test data for metrics")
        self.test_ds = test_ds

        if ctx_reqs is None:
            raise ValueError("Must pass ctx_reqs so they can be saved")
        self.ctx_reqs = ctx_reqs

        self.metrics = metrics
        self.tmp_checkpoint_dir = tmp_checkpoint_dir

    def __call__(self):
        from ray.air import session
        from ray.air.checkpoint import Checkpoint

        # Extract current step from dryml training process
        # step = self.train_state.global_step()

        temp_checkpoint_dir = os.path.join(
            self.tmp_checkpoint_dir,
            str(uuid.uuid4()))
        pathlib.Path(temp_checkpoint_dir).mkdir(exist_ok=False)
        # d = tempfile.TemporaryDirectory()
        # checkpoint_dir = d.name
        # Get checkpoint directory
        # with checkpoint.as_directory() as checkpoint_dir:

        # Create checkpoint dir
        #    if not os.path.exists(checkpoint_dir):
        #       # Create the checkpoint directory if needed
        #       pathlib.Path(checkpoint_dir).mkdir(parents=True, exist_ok=True)

        # Save model to checkpoint
        self.model.save_self(f"{temp_checkpoint_dir}/model.dry")
        self.train_state.save(f"{temp_checkpoint_dir}/train_state.pkl")
        with open(f"{temp_checkpoint_dir}/ctx_reqs.pkl", 'wb') as f:
            f.write(pickle.dumps(self.ctx_reqs))

        checkpoint = Checkpoint.from_directory(path=temp_checkpoint_dir)

        # Compute test set metrics
        metric_values = {}
        for metric_name in self.metrics:
            metric = self.metrics[metric_name]
            value = metric(self.model, self.test_ds)
            metric_values[metric_name] = value
        # Save model's dry_id
        metric_values['dry_id'] = self.model.dry_id

        # Report results
        session.report(metrics=metric_values, checkpoint=checkpoint)


class Tune2Trainer(object):
    def __init__(
            self,
            name=None,
            prep_method=None,
            metrics={}):
        self._name = name
        self.prep_method = prep_method
        self.metrics = metrics

    def __call__(self, config, checkpoint_dir=None):
        from ray.air import session

        prep_dict = self.prep_method()
        # Create repo
        repo = prep_dict['repo']()
        model_gen = prep_dict['model']
        data_gen = prep_dict['data']

        # Load or initialize training
        loaded_checkpoint = session.get_checkpoint()
        if loaded_checkpoint:
            with loaded_checkpoint.as_directory() as checkpoint_dir:
                # Load existing train state
                train_state = TrainSpec.load(
                    os.path.join(checkpoint_dir, "train_state.pkl"))

                # Load object from checkpoint
                model = dryml.load_object(
                    os.path.join(checkpoint_dir, "model.dry"),
                    repo=repo)

                # Retrieve requested context requirements
                ctx_filepath = os.path.join(checkpoint_dir, "ctx_reqs.pkl")
                if os.path.exists(ctx_filepath):
                    with open(ctx_filepath, 'rb') as f:
                        ctx_reqs = pickle.loads(f.read())
                else:
                    ctx_reqs = None

        else:
            # Initialize train spec
            train_state = TrainSpec()

            # Create build_obj function
            model_dict = model_gen(config, repo=repo)

            model = model_dict['model']
            if 'ctx_reqs' in model_dict:
                ctx_reqs = model_dict['ctx_reqs']
            else:
                ctx_reqs = model.dry_context_requirements()

        # Get data pipeline reqs
        if 'data_ctx' in prep_dict:
            data_ctx_reqs = prep_dict['data_ctx']()

            # Combine data and model contexts
            ctx_reqs = dryml.context.combine_reqs(
                ctx_reqs, data_ctx_reqs)

        # Acquire context
        dryml.context.set_context(ctx_reqs)

        # Start data pipelines
        data_dict = data_gen()
        train_ds = data_dict['train']
        test_ds = data_dict['test']

        # create train saver callable
        obj_saver = Tune2ObjectSaver(
            model=model,
            train_state=train_state,
            repo=repo,
            test_ds=test_ds,
            ctx_reqs=ctx_reqs,
            metrics=self.metrics,)
        callbacks = [obj_saver]

        # Prepare model for training
        model.prep_train()

        # Start training
        model.train(
            train_ds,
            train_spec=train_state,
            train_callbacks=callbacks)

        # Save trained model to repo
        repo.add_object(model)
        repo.save()

        # Save metric data
        final_metrics = {}
        for metric_name in self.metrics:
            metric = self.metrics[metric_name]
            metric_val = metric(model, test_ds)
            final_metrics[metric_name] = metric_val
        final_metrics.update(dry_id=model.dry_id, done=True)

        # Report metrics
        session.report(metrics=final_metrics)


class Tune1ObjectSaver(object):
    def __init__(
            self, model: Trainable = None,
            train_state: TrainSpec = None,
            repo: Repo = None,
            test_ds: Dataset = None,
            ctx_reqs=None,
            metrics={}):
        if model is None:
            raise ValueError("Must pass a model.")
        self.model = model

        if train_state is None:
            raise ValueError("Must pass a train state.")
        self.train_state = train_state

        if repo is None:
            raise ValueError("Must pass a repo.")
        self.repo = repo

        if test_ds is None:
            raise ValueError("Must pass test data for metrics")
        self.test_ds = test_ds

        if ctx_reqs is None:
            raise ValueError("Must pass ctx_reqs so they can be saved")
        self.ctx_reqs = ctx_reqs

        self.metrics = metrics

    def __call__(self):
        from ray import tune

        # Extract current step from dryml training process
        step = self.train_state.global_step()

        with tune.checkpoint_dir(step=step) as checkpoint_dir:
            if not os.path.exists(checkpoint_dir):
                # Create the checkpoint directory if needed
                pathlib.Path(checkpoint_dir).mkdir(parents=True, exist_ok=True)

            # Save model to checkpoint
            self.model.save_self(f"{checkpoint_dir}/model.dry")
            self.train_state.save(f"{checkpoint_dir}/train_state.pkl")
            with open(f"{checkpoint_dir}/ctx_reqs.pkl", 'wb') as f:
                f.write(pickle.dumps(self.ctx_reqs))

        # Compute test set metrics
        metric_values = {}
        for metric_name in self.metrics:
            metric = self.metrics[metric_name]
            value = metric(self.model, self.test_ds)
            metric_values[metric_name] = value
        # Save model's dry_id
        metric_values['dry_id'] = self.model.dry_id

        # Report results
        tune.report(**metric_values)


class Tune1Trainer(object):
    def __init__(
            self,
            name=None,
            prep_method=None,
            metrics={}):
        self._name = name
        self.prep_method = prep_method
        self.metrics = metrics

    def __call__(self, config, checkpoint_dir=None):
        from ray import tune

        prep_dict = self.prep_method()
        # Create repo
        repo = prep_dict['repo']()
        model_gen = prep_dict['model']
        data_gen = prep_dict['data']

        # Load or initialize training
        if checkpoint_dir:
            # Load existing train state
            train_state = TrainSpec.load(
                os.path.join(checkpoint_dir, "train_state.pkl"))

            # Load object from checkpoint
            model = dryml.load_object(
                os.path.join(checkpoint_dir, "model.dry"),
                repo=repo)

            # Retrieve requested context requirements
            ctx_filepath = os.path.join(checkpoint_dir, "ctx_reqs.pkl")
            if os.path.exists(ctx_filepath):
                with open(ctx_filepath, 'rb') as f:
                    ctx_reqs = pickle.loads(f.read())
            else:
                ctx_reqs = None

        else:
            # Initialize train spec
            train_state = TrainSpec()

            # Create build_obj function
            model_dict = model_gen(config, repo=repo)

            model = model_dict['model']
            if 'ctx_reqs' in model_dict:
                ctx_reqs = model_dict['ctx_reqs']
            else:
                ctx_reqs = model.dry_context_requirements()

        # Get data pipeline reqs
        if 'data_ctx' in prep_dict:
            data_ctx_reqs = prep_dict['data_ctx']()

            # Combine data and model contexts
            ctx_reqs = dryml.context.combine_reqs(
                ctx_reqs, data_ctx_reqs)

        # Acquire context
        dryml.context.set_context(ctx_reqs)

        # Start data pipelines
        data_dict = data_gen()
        train_ds = data_dict['train']
        test_ds = data_dict['test']

        # create train saver callable
        obj_saver = Tune1ObjectSaver(
            model=model,
            train_state=train_state,
            repo=repo,
            test_ds=test_ds,
            ctx_reqs=ctx_reqs,
            metrics=self.metrics,)
        callbacks = [obj_saver]

        # Prepare model for training
        model.prep_train()

        # Start training
        model.train(
            train_ds,
            train_spec=train_state,
            train_callbacks=callbacks)

        # Save trained model to repo
        repo.add_object(model)
        repo.save()

        # Save metric data
        final_metrics = {}
        for metric_name in self.metrics:
            metric = self.metrics[metric_name]
            metric_val = metric(model, test_ds)
            final_metrics[metric_name] = metric_val
        final_metrics.update(dry_id=model.dry_id, done=True)

        # Report metrics
        tune.report(**final_metrics)



from dryml.config import Meta
from dryml.models import Trainable
from dryml.data import Dataset
from dryml.data.torch import TorchDataset


class TorchDevice(Trainable):
    def __init__(self, device='cpu'):
        self.device = device
        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        pass

    def eval(self, data: Dataset, *args, **kwargs):
        if type(data) is not TorchDataset:
            raise TypeError("This trainable can only accept torch datasets.")

        return data.map_el(
            lambda el: el.to(self.device))


class TorchActivation(Trainable):
    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.act = None

        self.train_state = Trainable.trained

    def train(self, *args, train_spec=None, **kwargs):
        pass

    def compute_prepare_imp(self):
        self.act = self.cls(*self.args, **self.kwargs)

    def compute_cleanup_imp(self):
        del self.act
        self.act = None

    def eval(self, data: Dataset, *args, **kwargs):
        return data.apply_X(lambda X: self.act.forward(X))



from dryml.object import Object, ObjectFile, ObjectDef
from dryml.utils import is_nonstring_iterable, is_dictlike, get_class_str, \
    is_supported_scalar_type, is_supported_dictlike, is_supported_listlike, \
    map_dictlike, map_listlike
from typing import Union, Callable, Type, Mapping


# Need to recursively build selector when applied to a dictionary
def def_to_sel(val, cache=None):
    def applier(val):
        return def_to_sel(val, cache=cache)
    if is_supported_scalar_type(val):
        return val
    elif isinstance(val, Object):
        return Selector.from_def(val.definition(), cache=cache)
    elif isinstance(val, ObjectDef):
        return Selector.from_def(val, cache=cache)
    elif is_supported_listlike(val):
        return map_listlike(applier, val)
    elif is_supported_dictlike(val):
        return map_dictlike(applier, val)
    else:
        raise RuntimeError(
            f"Encountered unsupported value {val} of type {type(val)}")


class Selector(object):
    "Utility object for selecting for specific dry objects"

    @staticmethod
    def from_def(
            obj_def: ObjectDef,
            cache=None):

        # Create the cache if needed
        if cache is None:
            cache = {}

        # check the cache
        def_id = id(obj_def)
        if def_id in cache:
            return cache[def_id]

        new_args = def_to_sel(
            obj_def['dry_args'],
            cache=cache)

        new_kwargs = def_to_sel(
            obj_def['dry_kwargs'],
            cache=cache)

        sel = Selector(
            obj_def['cls'],
            args=new_args,
            kwargs=new_kwargs)

        cache[def_id] = sel

        return sel

    @staticmethod
    def from_dict(
            obj_dict: Mapping):
        raise RuntimeError("Functionality Questionable")
        args = obj_dict.get('dry_args', ())
        kwargs = obj_dict.get('dry_kwargs', {})
        return Selector(
            obj_dict['cls'],
            args=args,
            kwargs=kwargs)

    @staticmethod
    def from_obj(
            obj: Object):
        return Selector.from_def(
            obj.definition())

    @staticmethod
    def build(
            obj):
        if isinstance(obj, Object):
            return Selector.from_obj(obj)
        elif isinstance(obj, ObjectDef):
            return Selector.from_def(obj)
        elif isinstance(obj, Mapping):
            return Selector.from_dict(obj)
        else:
            raise TypeError(
                f"Can't construct Selector from type {type(obj)}")

    def __init__(self, cls: Type, args=None, kwargs=None):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    def match_objects(
            key_object, value_object, verbosity=0, cls_str_compare=True):
        if issubclass(type(key_object), type):
            # We have a type object. They match if
            res = issubclass(value_object, key_object)
            if not res and verbosity > 1:
                print(f"{value_object} is not a subclass of {key_object}")
            return res
        elif isinstance(key_object, Selector):
            res = key_object(
                value_object,
                verbosity=verbosity,
                cls_str_compare=cls_str_compare)
            if not res and verbosity > 1:
                print(f"{value_object} did not satisfy selector {key_object}")
            return res
        elif callable(key_object):
            res = key_object(value_object)
            if not res and verbosity > 1:
                print(f"callable on {value_object} failed.")
            return res
        elif isinstance(key_object, ObjectDef):
            return key_object.equal(
                value_object, cls_str_compare=cls_str_compare)
        elif isinstance(value_object, ObjectDef):
            return value_object.equal(
                key_object, cls_str_compare=cls_str_compare)
        elif is_dictlike(key_object):
            # dictlike branch is first because dictlike objects
            # are also iterable objects
            if not is_dictlike(value_object):
                if verbosity > 0:
                    print(f"a dict-like key object {key_object} not matched "
                          f"up with a dict-like value object {value_object}.")
                return False
            for key in key_object:
                if key not in value_object:
                    if verbosity > 0:
                        print(f"didn't find expected key {key} in "
                              f"{value_object}")
                    return False
                if not Selector.match_objects(
                        key_object[key],
                        value_object[key],
                        verbosity=verbosity,
                        cls_str_compare=cls_str_compare):
                    return False
        elif is_nonstring_iterable(key_object):
            if not is_nonstring_iterable(value_object):
                if verbosity > 0:
                    print(f"expected a nonstring iterable, got {value_object}")
                return False
            # For now, lists must have the same number of elements
            if len(key_object) != len(value_object):
                if verbosity > 0:
                    print("key object and value object have different "
                          f"lengths. expected {len(key_object)} got "
                          f"{len(value_object)}")
                return False
            for i in range(len(key_object)):
                # Each object must match
                if not Selector.match_objects(
                        key_object[i],
                        value_object[i],
                        verbosity=verbosity,
                        cls_str_compare=cls_str_compare):
                    return False
        else:
            if key_object != value_object:
                if verbosity > 0:
                    print(f"expected {key_object}, got {value_object}")
                return False

        return True

    def cls_compare(self, matcher, cls, verbosity=0, cls_str_compare=True):
        matched = True
        if isinstance(matcher, type):
            if not issubclass(cls, matcher):
                matched = False
        elif isinstance(matcher, Callable):
            if not matcher(cls):
                matched = False
        elif isinstance(matcher, str):
            if matcher != get_class_str(cls):
                matched = False
        else:
            raise ValueError(f"Unsupported matcher type: {type(matcher)}")
        if not matched:
            # Failure message
            if verbosity > 0:
                print("Class doesn't match")
            if verbosity > 1:
                print(f"Got {cls}, matcher: {matcher}")
            return False
        else:
            return True

    def args_compare(self, matcher, args, verbosity=0):
        if Selector.match_objects(matcher, args,
                                  verbosity=verbosity):
            return True
        else:
            if verbosity > 0:
                print("Args don't match")
            if verbosity > 1:
                print(f"Got {args}, matcher: {matcher}")
            return False

    def kwargs_compare(self, matcher, kwargs, verbosity=0):
        if Selector.match_objects(matcher, kwargs,
                                  verbosity=verbosity):
            return True
        else:
            if verbosity > 0:
                print("Kwargs don't match")
            if verbosity > 1:
                print(f"Got {kwargs}, matcher: {matcher}")
            return False

    def __call__(
            self,
            obj: Union[Object, ObjectFile, ObjectDef, Mapping],
            verbosity=0,
            cls_str_compare=True):
        if verbosity > 0:
            print(f"====== Selection started ({id(self)}) ({self.cls}) ======")
            if verbosity > 1:
                print(f"matching obj: {obj}")
        # Get definition
        if isinstance(obj, ObjectDef):
            obj_def = obj
        elif isinstance(obj, Object) or isinstance(obj, ObjectFile):
            obj_def = obj.definition()
        elif isinstance(obj, Mapping):
            raise RuntimeError("Not currently supported")
        else:
            if verbosity > 0:
                print(
                    f"input object not a matchable object. type: {type(obj)}")
            return False

        # If required, check object class
        if self.cls is not None:
            if not self.cls_compare(
                    self.cls,
                    obj_def.cls,
                    verbosity=verbosity,
                    cls_str_compare=cls_str_compare):
                if verbosity > 0:
                    print("Class didn't match")
                if verbosity > 1:
                    print(f"Expected class {self.cls} got {obj_def.cls}")
                return False

        # Check object args
        if self.args is not None:
            if not self.args_compare(
                    self.args,
                    obj_def.args,
                    verbosity=verbosity):
                if verbosity > 0:
                    print("Args didn't match")
                if verbosity > 1:
                    print(f"Expected args {self.args} got {obj_def.args}")
                return False

        # Check object kwargs
        if self.kwargs is not None:
            if not self.kwargs_compare(
                    self.kwargs,
                    obj_def.kwargs,
                    verbosity=verbosity):
                if verbosity > 0:
                    print("Kwargs didn't match")
                if verbosity > 1:
                    print(f"Expected kwargs {self.kwargs} "
                          f"got {obj_def.kwargs}")
                return False

        return True

    def __str__(self):
        return f"Selector({self.cls}, {self.args}, {self.kwargs})"

    def repr(self):
        return f"{self}"



from dryml import Object, Meta
from dryml.models import Trainable, Component
from dryml.models import TrainFunction as BaseTrainFunction
from dryml.data import Dataset


class Wrapper(Object):
    __dry_compute_context__ = 'tf'

    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, *args, **kwargs):
        if not isinstance(cls, type):
            raise TypeError(
                f"Expected first argument to be type. Got {type(cls)}")
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.obj = None

    def compute_prepare_imp(self):
        self.obj = self.cls(*self.args, **self.kwargs)

    def compute_cleanup_imp(self):
        del self.obj
        self.obj = None


class TrainFunction(BaseTrainFunction):
    __dry_compute_context__ = 'tf'


class Model(Component):
    __dry_compute_context__ = 'tf'


class Trainable(Trainable):
    __dry_compute_context__ = 'tf'

    def eval(self, data: Dataset, *args, eval_batch_size=32, **kwargs):
        def eval_func(X):
            return self.model(X, *args, **kwargs)
        if data.batched:
            # We can execute the method directly on the data
            return data.tf() \
                       .apply_X(func=eval_func)
        else:
            # We first need to batch the data, then unbatch to leave
            # The dataset character unchanged.
            return data.tf() \
                       .batch(batch_size=eval_batch_size) \
                       .apply_X(func=eval_func) \
                       .unbatch()



import os
import shutil
import pathlib
import collections
import inspect
import hashlib
import importlib
import pickle
from typing import Type, Union, IO, Optional, Callable
import zipfile
import io
import numpy as np


def is_in_typelist(val, typelist):
    for t in typelist:
        if isinstance(val, t):
            return True
    return False


supported_scalar_types = (
    str, bytes, int, float, bool, type,
    np.number, np.bool_)


def is_supported_scalar_type(val):
    if val is None:
        return True
    return is_in_typelist(val, supported_scalar_types)


supported_listlike_types = (
    tuple, list)


def is_supported_listlike(val):
    return is_in_typelist(val, supported_listlike_types)


def map_listlike(func, val):
    the_type = type(val)
    return the_type(map(func, val))


supported_dictlike_types = (
    dict,)


def is_supported_dictlike(val):
    return is_in_typelist(val, supported_dictlike_types)


def map_dictlike(func, val):
    the_type = type(val)
    return the_type({
        k: func(val[k]) for k in val})


def equal_listlike(equal_func, list_a, list_b):
    if len(list_a) != len(list_b):
        return False

    for i in range(len(list_a)):
        el_a = list_a[i]
        el_b = list_b[i]
        if not equal_func(el_a, el_b):
            return False

    return True


def equal_dictlike(equal_func, dict_a, dict_b):
    keys_a = set(dict_a.keys())
    keys_b = set(dict_b.keys())
    if keys_a != keys_b:
        return False

    for key in keys_a:
        val_a = dict_a[key]
        val_b = dict_b[key]
        if not equal_func(val_a, val_b):
            return False

    return True


def is_nonstring_iterable(val):
    if isinstance(val, collections.abc.Iterable) and type(val) \
             not in [str, bytes]:
        return True
    else:
        return False


def is_dictlike(val):
    return isinstance(val, collections.abc.Mapping)


def is_iterator(val):
    # Implementation from https://stackoverflow.com/a/36230057
    if (
            hasattr(val, '__iter__') and
            hasattr(val, '__next__') and
            callable(val.__iter__) and
            val.__iter__() is val):
        return True
    else:
        return False


def init_arg_list_handler(arg_list):
    if arg_list is None:
        return []
    else:
        return arg_list


def init_arg_dict_handler(arg_dict):
    if arg_dict is None:
        return {}
    else:
        return arg_dict


def get_class_str(obj):
    if isinstance(obj, type):
        return '.'.join([inspect.getmodule(obj).__name__,
                         obj.__name__])
    else:
        return '.'.join([inspect.getmodule(obj).__name__,
                         obj.__class__.__name__])


def get_class_by_name(module: str, cls: str, reload: bool = False):
    module = importlib.import_module(module)
    # If indicated, reload the module.
    if reload:
        module = importlib.reload(module)
    return getattr(module, cls)


def get_class_from_str(cls_str: str, reload: bool = False):
    cls_split = cls_str.split('.')
    module_string = '.'.join(cls_split[:-1])
    cls_name = cls_split[-1]
    return get_class_by_name(module_string, cls_name, reload=reload)


def get_current_cls(cls: Type, reload: bool = False):
    return get_class_by_name(
        inspect.getmodule(cls).__name__,
        cls.__name__, reload=reload)


def get_hashed_id(hashstr: str):
    return hashlib.md5(hashstr.encode('utf-8')).hexdigest()


def path_needs_directory(path):
    """
    Method to determine whether a path is absolute, relative, or just
    a plain filename. If it's a plain filename, it needs a directory
    """
    head, tail = os.path.split(path)
    if head == '':
        return True
    else:
        return False


def pickler(obj):
    "Method to ensure all objects are pickled in the same way"
    # Consider updating to protocol=5 when python 3.7 is deprecated
    return pickle.dumps(obj, protocol=4)


def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate


def show_sig(sig):
    for key in sig.parameters:
        par = sig.parameters[key]
        print(f"{par.name} - {par.kind} - {par.default}")


def adjust_class_module(cls):
    # Set module properly
    # From https://stackoverflow.com/questions/1095543/
    #              get-name-of-calling-functions-module-in-python
    # We go up two functions, one to get to the calling function,
    # Another to get to that function's caller. That should be
    # in a module.
    frm = inspect.stack()[2]
    calling_mod = inspect.getmodule(frm[0])
    cls.__module__ = calling_mod.__name__


def get_temp_checkpoint_dir(dry_id):
    home_dir = os.environ['HOME']
    # Create checkpoint dir
    temp_checkpoint_dir = os.path.join(home_dir, '.dryml', dry_id)
    # Create directory if needed
    pathlib.Path(temp_checkpoint_dir).mkdir(parents=True, exist_ok=True)
    return temp_checkpoint_dir


def cleanup_checkpoint_dir(checkpoint_dir):
    shutil.rmtree(checkpoint_dir)


def head(get_result):
    from dryml import Object
    if issubclass(type(get_result), Object):
        return get_result
    return get_result[0]


def tail(get_result):
    from dryml import Object
    if issubclass(type(get_result), Object):
        return get_result
    return get_result[-1]


def count(get_result):
    from dryml import Object
    from dryml.repo import RepoContainer
    if isinstance(get_result, Object):
        return 1
    elif isinstance(get_result, RepoContainer):
        return 1
    return len(get_result)


def show_contained_objects(save_file: Union[str, IO[bytes]]):
    if type(save_file) is str or IO[bytes]:
        zf = zipfile.ZipFile(save_file, mode='r')
    else:
        zf = save_file

    print(zf.namelist())


def show_contained_objects_md5(save_file: Union[str, IO[bytes]]):
    close_file = False
    if type(save_file) is zipfile.ZipFile:
        zf = save_file
    else:
        zf = zipfile.ZipFile(save_file, mode='r')
        close_file = True

    for name in zf.namelist():
        with zf.open(name, 'r') as f:
            m = hashlib.md5()
            m.update(f.read())
            print(f"{name}: {m.hexdigest()}")

    if close_file:
        zf.close()


file_blocklist = [
   'meta_data.pkl',
   'cls_def.dill',
   'dry_args.pkl',
   'dry_kwargs.pkl',
   'dry_mut.pkl',
]


def create_zip_branch(input_file, tag):
    content_nodes = {}
    try:
        with zipfile.ZipFile(input_file, mode='r') as zf:
            content_names = zf.namelist()
            for name in content_names:
                if name[-4:] == '.zip':
                    with zf.open(name, mode='r') as f:
                        content_nodes = {
                            **content_nodes,
                            **create_zip_branch(f, name)}
                else:
                    content_nodes[name] = {}
    except zipfile.BadZipFile:
        tag = f"{tag} - Bad Zipfile"
    return {tag: content_nodes}


def create_object_tree_from_dryfile(input_file, tag, report_class=True):
    obj_def = None
    content_nodes = {}
    try:
        import dryml
        with dryml.ObjectFile(input_file) as dry_f:
            def_problem = False
            try:
                obj_def = dry_f.definition()
            except KeyError as e:
                def_problem = True
                tag = f"{tag} - Issue reading object definition ({e})"

            if not def_problem:
                content_names = dry_f.z_file.namelist()
                for name in content_names:
                    if name[-4:] == '.dry':
                        # this is another dry file.
                        with dry_f.z_file.open(name, mode='r') as f:
                            content_nodes = {
                                **content_nodes,
                                **create_object_tree_from_dryfile(
                                    f,
                                    name,
                                    report_class=report_class)}

                    elif name not in file_blocklist:
                        if name[-4:] == '.zip':
                            with dry_f.z_file.open(name, mode='r') as f:
                                content_nodes = {
                                    **content_nodes,
                                    **create_zip_branch(f, name)}
                        else:
                            content_nodes[name] = {}
    except zipfile.BadZipFile:
        tag = f"{tag} - Bad Zip File"

    if tag is not None:
        if obj_def is not None:
            label = f"{tag}: {obj_def.dry_id}"
        else:
            label = f"{tag}"
    else:
        if obj_def is not None:
            label = f"{obj_def.dry_id}"
        else:
            label = "None"

    if report_class and obj_def is not None:
        label += f" ({dryml.utils.get_class_str(obj_def.cls)})"

    if len(content_nodes) > 0:
        return {label: content_nodes}
    else:
        return {label: {}}


def create_file_tree_from_zipfile(input_file, tag):
    content_nodes = {}
    try:
        with zipfile.ZipFile(input_file) as zf:
            content_names = zf.namelist()
            for name in content_names:
                if name[-4:] == '.dry' or name[-4:] == '.zip':
                    # this is another zip file.
                    with zf.open(name, mode='r') as f:
                        content_nodes = {
                            **content_nodes,
                            **create_zip_branch(
                                f,
                                name)}

                else:
                    content_nodes[name] = {}
    except zipfile.BadZipFile:
        tag = f"{tag} - Bad Zip File"

    label = f"{tag}"

    if len(content_nodes) > 0:
        return {label: content_nodes}
    else:
        return {label: {}}


def show_object_tree_from_dryfile(input_file, report_class=True):
    cur_pos = None
    if isinstance(input_file, io.IOBase):
        try:
            # Get current file's position, and rewind to beginning
            cur_pos = input_file.tell()
            input_file.seek(0)
        except io.UnsupportedOperation:
            print(
                "Can't show file content of file of type "
                f"{type(input_file)}. seeking not supported.")
            return
    dry_tree = create_object_tree_from_dryfile(
        input_file,
        None,
        report_class=report_class)
    from asciitree import LeftAligned
    from asciitree.drawing import BoxStyle, BOX_LIGHT
    tr = LeftAligned(
        draw=BoxStyle(
            gfx=BOX_LIGHT,
            horiz_len=1
        )
    )
    print(tr(dry_tree))
    if isinstance(input_file, io.IOBase):
        # Go back to where we were in the file.
        input_file.seek(cur_pos)


def show_files_from_zipfile(input_file):
    cur_pos = None
    if isinstance(input_file, io.IOBase):
        try:
            # Save current file position and go to beginning of file
            cur_pos = input_file.tell()
            input_file.seek(0)
        except io.UnsupportedOperation:
            print(
                f"Can't show content of file {input_file}. "
                "seeking not supported.")
            return
    zip_tree = create_file_tree_from_zipfile(
        input_file,
        None)
    from asciitree import LeftAligned
    from asciitree.drawing import BoxStyle, BOX_LIGHT
    tr = LeftAligned(
        draw=BoxStyle(
            gfx=BOX_LIGHT,
            horiz_len=1
        )
    )
    print(tr(zip_tree))
    if isinstance(input_file, io.IOBase):
        # Return file to original position
        input_file.seek(cur_pos)


def apply_func(
        obj, func, func_args=None, sel=Optional[Callable],
        func_kwargs=None):
    if func_args is None:
        func_args = ()
    if func_kwargs is None:
        func_kwargs = {}

    for sub_obj in obj.__dry_obj_container_list__:
        apply_func(
            sub_obj, func, func_args=func_args, sel=sel,
            func_kwargs=func_kwargs)

    if sel is None or sel(obj):
        func(obj, *func_args, **func_kwargs)



import pytest
import dryml


@pytest.mark.usefixtures("create_name")
def test_basic_trainable_1(create_name):
    import objects

    test_obj = objects.HelloTrainable(msg='test1')

    test_obj.save_self(create_name)

    test_obj2 = dryml.load_object(create_name)

    assert test_obj.definition().get_individual_id() == \
        test_obj2.definition().get_individual_id()

    assert test_obj2.dry_kwargs['msg'] == 'test1'


def test_basic_trainable_2():
    import objects

    test_obj = objects.HelloTrainable(msg='test1')

    new_obj = dryml.change_object_cls(
        test_obj, cls=objects.HelloTrainable, update=True)

    assert test_obj.definition() == new_obj.definition()


def test_trainable_def_1():
    import objects

    test_obj = objects.HelloTrainableC('test1')

    test_obj2 = test_obj.definition().build()

    assert test_obj.definition() == test_obj2.definition()
    assert test_obj.dry_args[0] == test_obj2.dry_args[0]


def test_trainable_def_2():
    import objects

    test_obj = objects.HelloTrainableC(objects.HelloStr(msg='test'))

    test_obj2 = test_obj.definition().build()

    assert test_obj.definition() == test_obj2.definition()
    assert type(test_obj.A) is objects.HelloStr
    assert type(test_obj2.A) is objects.HelloStr
    assert test_obj.dry_args[0].str_message == \
        test_obj2.dry_args[0].str_message
    assert test_obj.A.definition() == test_obj2.A.definition()
    assert test_obj.A.str_message == test_obj2.A.str_message


def test_trainable_def_3():
    import objects

    test_obj = objects.HelloTrainableD(A='test1')

    assert test_obj.A == 'test1'

    test_obj2 = test_obj.definition().build()

    assert test_obj.definition() == test_obj2.definition()
    assert test_obj.dry_kwargs['A'] == 'test1'
    assert test_obj2.dry_kwargs['A'] == 'test1'
    assert type(test_obj.A) is str
    assert type(test_obj2.A) is str
    assert test_obj.A == test_obj2.A


def test_trainable_def_4():
    import objects

    test_obj = objects.HelloTrainableD(A=objects.HelloStr(msg='test'))

    test_obj2 = test_obj.definition().build()

    assert test_obj.definition() == test_obj2.definition()
    assert type(test_obj.A) is objects.HelloStr
    assert type(test_obj2.A) is objects.HelloStr
    assert test_obj.dry_kwargs['A'].str_message == \
        test_obj2.dry_kwargs['A'].str_message
    assert test_obj.A.definition() == test_obj2.A.definition()
    assert test_obj.A.str_message == test_obj2.A.str_message



import numpy as np
import pytest
try:
    import tensorflow as tf
except ImportError:
    pytest.skip(allow_module_level=True)

from dryml.data.tf import TFDataset
from dryml.data import NumpyDataset
import dryml
from utils import ray_wrap


@ray_wrap
def test_tf_dataset_1():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 10

        data = np.random.random((batch_size, 20))

        tf_ds = tf.data.Dataset.from_tensor_slices(data)

        dataset = TFDataset(tf_ds)

        # assert len(dataset) == batch_size

        i = 0
        for el in dataset.unbatch():
            assert np.all(el.numpy() == data[i])
            i += 1


@ray_wrap
def test_tf_dataset_2():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 10

        data = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data)

        dataset = TFDataset(tf_data)

        # assert len(dataset) == batch_size

        dataset = dataset.batch(batch_size=batch_size)

        # assert len(dataset) == batch_size

        v = next(iter(dataset))

        assert np.all(v.numpy() == data)


@ray_wrap
def test_tf_dataset_3():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data, batch_size=len(batch_rows))

        i = 0
        for el in dataset.unbatch():
            assert np.all(el.numpy() == batch_rows[i])
            i += 1


@ray_wrap
def test_tf_dataset_4():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data, batch_size=len(batch_rows))

        v = next(iter(dataset))

        all_data = np.stack(batch_rows, axis=0)

        assert np.all(v.numpy() == all_data)


@ray_wrap
def test_tf_dataset_5():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data, batch_size=len(batch_rows))

        sub_batch_size = batch_size//4

        i = 0
        for el in dataset.batch(batch_size=sub_batch_size):
            sub_batch = np.stack(
                batch_rows[(i*sub_batch_size):((i+1)*sub_batch_size)], axis=0)
            assert np.all(el.numpy() == sub_batch)
            i += 1


@ray_wrap
def test_tf_dataset_6():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data, batch_size=len(batch_rows)).unbatch()

        dataset2 = dataset.batch(batch_size//4).unbatch()

        for el, el2 in zip(dataset, dataset2):
            assert np.all(el.numpy() == el2.numpy())


@ray_wrap
def test_tf_dataset_7():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data, batch_size=len(batch_rows)).unbatch()

        i = 0
        for idx, el in dataset.as_indexed():
            assert idx.numpy() == i
            assert np.all(el.numpy() == batch_rows[i])
            i += 1


@ray_wrap
def tf_dataset_8():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        data = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data)

        dataset = TFDataset(tf_data)

        i = 0
        for idx, el in dataset.as_indexed():
            assert idx.numpy() == i
            assert np.all(el.numpy() == data[i])
            i += 1


@ray_wrap
def test_tf_dataset_9():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        data = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data)

        dataset = TFDataset(tf_data)

        i = 0
        for el in dataset.as_indexed().as_not_indexed():
            assert np.all(el.numpy() == data[i])
            i += 1


@ray_wrap
def test_tf_dataset_10():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        data1 = np.random.random((batch_size, 20))
        data2 = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices((data1, data2))

        dataset = TFDataset(tf_data)

        i = 0
        for idx, el in dataset.as_indexed():
            assert idx.numpy() == i
            assert np.all(el[0].numpy() == data1[i])
            assert np.all(el[1].numpy() == data2[i])
            i += 1
        assert i == batch_size


@ray_wrap
def test_tf_dataset_11():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        data1 = np.random.random((batch_size, 20))
        data2 = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices((data1, data2))

        dataset = TFDataset(tf_data).batch(batch_size=32)

        assert dataset is not dataset.batch(batch_size=2)


@ray_wrap
def test_tf_dataset_12():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32

        batch_rows = []
        for i in range(batch_size):
            batch_rows.append(np.random.random((20,)))

        tf_data = tf.data.Dataset.from_tensors(batch_rows)

        dataset = TFDataset(tf_data).batch(batch_size=batch_size)

        assert dataset is dataset.batch(batch_size=batch_size)


@ray_wrap
def test_tf_dataset_13():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset = TFDataset(tf_data).batch(batch_size=batch_size)

        take_num = 10
        dataset = dataset.unbatch().take(take_num)

        count = 0
        for el in dataset:
            count += 1
            assert el.shape == (20,)

        assert count == take_num


@ray_wrap
def test_tf_dataset_14():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset = TFDataset(tf_data)

        el1 = dataset.peek()
        el2 = dataset.peek()

        assert np.all(el1.numpy() == el2.numpy())


@ray_wrap
def test_tf_dataset_15():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset = TFDataset(tf_data)

        for el_a, el_b in zip(dataset, dataset):
            assert np.all(el_a.numpy() == el_b.numpy())


@ray_wrap
def test_tf_to_numpy_dataset_1():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 20))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset = TFDataset(tf_data)

        numpy_dataset = dataset.numpy()
        i = 0
        for tf_el, numpy_el in zip(dataset, numpy_dataset):
            assert np.all(tf_el.numpy() == numpy_el)
            i += 1
        assert i == batch_size


@ray_wrap
def test_numpy_to_tf_dataset_2():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 20))

        dataset = NumpyDataset(data_block, batch_size=batch_size)

        tf_dataset = dataset.tf()

        numpy_data = dataset.peek()
        tf_data = tf_dataset.peek()

        assert np.all(numpy_data == tf_data.numpy())


@ray_wrap
def test_chain_transforms_1():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        dataset = NumpyDataset(data_block, batch_size=batch_size)

        dataset = dataset.unbatch().tf()

        i = 0
        for el_a, el_b in zip(dataset, dataset):
            assert np.all(el_a.numpy() == data_block[i])
            assert np.all(el_b.numpy() == data_block[i])
            assert np.all(el_a.numpy() == el_b.numpy())
            i += 1


@ray_wrap
def test_chain_transforms_1a():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        dataset = NumpyDataset(data_block, batch_size=batch_size)

        dataset = dataset.unbatch().tf()

        it1 = iter(dataset)
        it2 = iter(dataset)

        el_a = next(it1)
        el_b = next(it2)

        assert np.all(el_a.numpy() == data_block[0])
        assert np.all(el_b.numpy() == data_block[0])
        assert np.all(el_a.numpy() == el_b.numpy())


@ray_wrap
def test_chain_transforms_2():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset = TFDataset(tf_data)

        dataset = dataset.numpy()

        i = 0
        for el_a, el_b in zip(dataset, dataset):
            assert np.all(el_a == data_block[i])
            assert np.all(el_b == data_block[i])
            assert np.all(el_a == el_b)
            i += 1


@ray_wrap
def test_chain_transforms_3():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset_orig = TFDataset(tf_data)

        dataset = dataset_orig.numpy().tf()

        i = 0
        for el1, el2 in zip(dataset_orig, dataset):
            assert np.all(el1.numpy() == el2.numpy())
            i += 1
        assert i == batch_size


@ray_wrap
def test_chain_transforms_4():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        tf_data = tf.data.Dataset.from_tensor_slices(data_block)

        dataset_orig = TFDataset(tf_data)

        dataset = dataset_orig.numpy().apply_X(lambda X: X**2).tf()

        i = 0
        for el1, el2 in zip(dataset_orig, dataset):
            assert np.all(el1.numpy()**2 == el2.numpy())
            i += 1
        assert i == batch_size


@ray_wrap
def test_chain_transforms_5():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block_x = np.random.random((batch_size, 5))
        data_block_y = np.random.random((batch_size, 5))

        tf_data = tf.data.Dataset.from_tensor_slices(
            (data_block_x, data_block_y))

        dataset_orig = TFDataset(tf_data, supervised=True)

        dataset = dataset_orig.numpy().apply_X(lambda X: X**2).tf()

        i = 0
        for el1, el2 in zip(dataset_orig, dataset):
            assert np.all(el1[0].numpy()**2 == el2[0].numpy())
            assert np.all(el1[1].numpy() == el2[1].numpy())
            i += 1
        i == batch_size


@ray_wrap
def test_chain_transforms_6():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        dataset = NumpyDataset((data_block, data_block), supervised=True)

        dataset = dataset.tf().unbatch().numpy()

        for e1, e2 in dataset:
            assert np.all(e1 == e2)


@ray_wrap
def test_chain_transforms_8():
    with dryml.context.ContextManager({'tf': {}}):
        batch_size = 32
        data_block = np.random.random((batch_size, 5))

        dataset = NumpyDataset(data_block, supervised=False)

        dataset = dataset.tf().numpy()

        dataset_peek = dataset.peek()
        assert type(dataset_peek) is np.ndarray
        assert np.all(data_block == dataset_peek)



"""
Code to run subprocesses with proper contexts
"""


from dryml.context.context_tracker import contexts, WrongContextError, \
    context, consolidate_contexts, get_context_requirements, \
    ContextManager, NoContextError
import copy
import multiprocessing as mp
import traceback
import functools
import io
import zipfile
import time
import dill
from dryml.save_cache import SaveCache

mp_ctx = mp.get_context('spawn')


# Wrapping process base starting from
# https://stackoverflow.com/a/33599967/2485932
# This ensures the process will exit even if an error
# is thrown.
class Process(mp_ctx.Process):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pconn, self._cconn = mp_ctx.Pipe()
        self._exception = None

    def run(self):
        try:
            super().run()
            self._cconn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


def cls_method_compute(fn_name, **ctx_kwargs):
    """
    Mark a class method as compute. DryMeta will ensure
    such named methods are wrapped in a compute context.
    """

    def _dec(cls):
        cls_method_list = getattr(cls, '__dry_compute_methods__', [])
        cls_method_list.append((fn_name, ctx_kwargs))
        cls.__dry_compute_methods__ = cls_method_list
        return cls

    return _dec


# Get list of dry objects because we need to find a good
# context to use.
def get_dry_objects(*args, **kwargs):
    from dryml import Object

    dry_objects = []
    for arg in args:
        if isinstance(arg, Object):
            dry_objects.append(arg)
    for name in kwargs:
        arg = kwargs[name]
        if isinstance(arg, Object):
            dry_objects.append(arg)
    return dry_objects


# extra methods for activating/deactivating objects
def activate_objects(obj_list):
    for obj in obj_list:
        obj.compute_activate()


class process_executor(object):
    def __init__(
            self,
            f=None,
            ctx_reqs={},
            update_obj_defs=[],
            verbose=False,
            args=[],
            kwargs={}):
        self.f_ser = dill.dumps(f)
        self.ctx_reqs = ctx_reqs
        self.update_obj_defs = update_obj_defs
        self.verbose = verbose
        self.args_ser = dill.dumps(args)
        self.kwargs_ser = dill.dumps(kwargs)

    def final_call(
            self,
            f,
            ctx_send_q,
            ctx_ret_q,
            *args,
            **kwargs):

        # import resource
        # max_mem = 4*1024*1024*1024
        # resource.setrlimit(
        #     resource.RLIMIT_AS,
        #     (max_mem, resource.RLIM_INFINITY))

        if ctx_send_q is None:
            raise RuntimeError("A send queue is required.")

        if ctx_ret_q is None:
            raise RuntimeError("A return queue is required.")

        # Activate context
        with ContextManager(resource_requests=self.ctx_reqs) as ctx_mgr:
            # Get list of dry_objects
            dry_objects = get_dry_objects(*args, **kwargs)

            # Activate unactivated objects
            activate_objects(dry_objects)

            # Execute method
            res = f(*args, **kwargs)

            # Put result in return queue
            ctx_ret_q.put(res)

            save_cache = SaveCache()

            # Put object updates in queue
            if len(self.update_obj_defs) > 0:
                for obj_def in self.update_obj_defs:
                    found = False
                    for obj in dry_objects:
                        if obj_def == obj.definition():
                            res_buf = io.BytesIO()
                            obj.save_self(res_buf, save_cache=save_cache)
                            res_buf.seek(0)
                            ctx_ret_q.put(res_buf.read())
                            found = True
                            break
                    # Close int_files in save_cache
                    if not found:
                        raise RuntimeError(
                            "Couldn't find an object to associate"
                            " a definition to!")

            ctx_mgr.deactivate_objects(save_cache=save_cache)

            # Wait until return queue is empty (parent thread has emptied it)
            while not ctx_ret_q.empty():
                # Sleep a bit
                time.sleep(0.1)

        # Delete save cache
        del save_cache

    def __call__(self, ctx_send_q, ctx_ret_q):
        if self.verbose:
            print(
                f"{mp.current_process().pid}: New process "
                f"started pid:{mp.current_process().pid}")
        # Undill function
        f = dill.loads(self.f_ser)

        # Undill args/kwargs
        args = dill.loads(self.args_ser)
        kwargs = dill.loads(self.kwargs_ser)

        ph_data = ctx_send_q.get()

        from dryml.object import reconstruct_args_kwargs
        if self.verbose:
            print(
                f"{mp.current_process().pid}: Reconstructing arguments: "
                f"args: {args} kwargs {kwargs}")
        reconstruct_args_kwargs(
            args, kwargs, ph_data, verbose=self.verbose)

        self.final_call(f, ctx_send_q, ctx_ret_q, *args, **kwargs)


def compute_context(
        ctx_context_reqs=None,
        ctx_use_existing_context=True,
        ctx_dont_create_context=False,
        ctx_update_objs=False,
        ctx_verbose=False):

    """
        ctx_context_reqs: if None, will try to guess
        ctx_use_existing_context: if True, will try to run in
            the currently active context without a subprocess
        ctx_dont_create_context: if True, will avoid trying to create
            a context, instead relying on the existing context.
        ctx_update_objs: if True, will serialize objects after method is run
            and update local objects by running load_object with
            the serialized results
    """

    def _func_dec(f):
        if hasattr(f, '__dry_context_wrapped__'):
            # Don't wrap a function twice
            if f.__dry_context_wrapped__:
                return f

        @functools.wraps(f)
        def wrapped_func(
                *args,
                call_context_reqs=None,
                call_use_existing_context=None,
                call_dont_create_context=None,
                call_update_objs=None,
                call_update_skiplist=None,
                call_verbose=None,
                **kwargs):
            """
                call_*: call version of the ctx variables which take
                    precedence.
                call_update_skiplist: A list of objects to skip
                    when updating objects.
            """

            # Check whether we are using an existing context
            use_existing_context = ctx_use_existing_context
            if call_use_existing_context is not None:
                use_existing_context = call_use_existing_context
            dont_create_context = ctx_dont_create_context
            if call_dont_create_context is not None:
                dont_create_context = call_dont_create_context

            verbose = ctx_verbose
            if call_verbose is not None:
                verbose = call_verbose

            ctx_reqs = ctx_context_reqs
            if call_context_reqs is not None:
                ctx_reqs = call_context_reqs

            if ctx_reqs is None:
                # Determine needed contexts
                ctx_reqs = get_context_requirements(get_dry_objects(
                    *args, **kwargs))

            if use_existing_context:
                ctx_manager = context()
                if ctx_manager is not None:
                    if not ctx_manager.satisfies(ctx_reqs):
                        raise WrongContextError(
                            "Current context is not appropriate")
                else:
                    # No currently active context, need to create one
                    use_existing_context = False

            if use_existing_context:
                # Execute the method in this thread.
                # Activate objects which don't have a context active.
                activate_objects(get_dry_objects(
                    *args, **kwargs))

                # Execute method
                res = f(*args, **kwargs)

                # We don't have to wory about any of the
                # context management stuff

                return res
            else:
                if dont_create_context:
                    raise NoContextError("Instructed to not create a context!")

                # Execute the method in another thread.
                # get a list of objects we will need to update.
                update_objs = False
                if call_update_objs is not None:
                    update_objs = call_update_objs
                elif ctx_update_objs:
                    update_objs = True

                # record which objects to update
                update_objs_list = []
                update_obj_defs = []
                if update_objs:
                    # build update skiplist
                    if call_update_skiplist is None:
                        update_skiplist = []
                    else:
                        update_skiplist = call_update_skiplist

                    dry_objects = get_dry_objects(*args, **kwargs)

                    # Get list of dry objects
                    update_objs_list = list(filter(
                        lambda o: o not in update_skiplist,
                        dry_objects))

                    # Translate direct list into definitions
                    update_obj_defs = list(map(
                        lambda o: o.definition(),
                        update_objs_list))

                ctx_ret_q = mp_ctx.Queue()
                ctx_send_q = mp_ctx.Queue()

                from dryml.object import prep_args_kwargs

                # Replace Objects in args/kwargs with placeholders
                # Get placeholder data
                (args, kwargs), ph_data = prep_args_kwargs(args, kwargs)

                executor = process_executor(
                    f=f, ctx_reqs=ctx_reqs,
                    update_obj_defs=update_obj_defs,
                    verbose=verbose,
                    args=args,
                    kwargs=kwargs)

                try:
                    from pytest_cov.embed import cleanup_on_sigterm
                except ImportError:
                    pass
                else:
                    cleanup_on_sigterm()

                # run function
                p = Process(target=executor, args=[ctx_send_q, ctx_ret_q])
                p.start()

                # Send placeholder data
                ctx_send_q.put(ph_data)

                sub_pid = p.pid

                if verbose:
                    print(f"Started context isolation process, pid: {sub_pid}")

                # Check loop
                queue_results = []

                # Num elements which need to be fetched before joining
                num_to_fetch = 1
                num_to_fetch += len(update_objs_list)

                def check_queue():
                    while len(queue_results) < num_to_fetch:
                        # Check whether the queue is empty
                        if not ctx_ret_q.empty():
                            # Get result
                            queue_results.append(ctx_ret_q.get())
                        else:
                            break

                while p.is_alive():
                    # check for exception
                    if p.exception is not None:
                        e, tb = p.exception
                        print(
                            "Exception encountered in context "
                            f"thread! pid: {p.pid}")
                        print(tb)
                        # rejoin thread
                        p.join()
                        # Reraise
                        raise e

                    # Get results, we need to do this before the join
                    # to prevent deadlock from large data being shuttled
                    # through the queue.
                    check_queue()

                    # Sleep a bit
                    time.sleep(0.1)

                if p.exception is not None:
                    e, tb = p.exception
                    print("Exception encountered in context "
                          f"thread! pid: {p.pid}")
                    print(tb)
                    # rejoin thread
                    p.join()
                    # Reraise
                    raise e

                # Check queue again
                check_queue()

                # join the thread
                p.join()

                if len(queue_results) < num_to_fetch:
                    raise RuntimeError("Didn't get all expected results!")

                # retrieve return value
                retval = queue_results.pop(0)

                # Update dry objects
                load_issue_list = []
                for obj in update_objs_list:
                    obj_buf = queue_results.pop(0)
                    obj_buf = io.BytesIO(obj_buf)
                    from dryml.object import load_object_content
                    if not load_object_content(obj, obj_buf):
                        load_issue_list.append(obj)

                # Close thread
                p.close()

                # Close and Delete the temporary queue
                ctx_ret_q.close()
                del ctx_ret_q

                if len(load_issue_list):
                    print("There was an issue updating the following objects")
                    for obj in load_issue_list:
                        print(obj)

                if verbose:
                    print(f"Ended context isolation process, pid: {sub_pid}")

                return retval

        wrapped_func.__dry_context_wrapped__ = True

        # Return wrapped function
        return wrapped_func

    return _func_dec


def compute(f):
    return compute_context()(f)


class tune_process_executor(object):
    def __init__(
            self,
            f=None,
            ctx_name=None,
            context_kwargs={},
            update_obj_defs=[],
            args=[],
            kwargs={}):
        self.f_ser = dill.dumps(f)
        self.ctx_name = ctx_name
        self.context_kwargs = context_kwargs
        self.update_obj_defs = update_obj_defs
        self.args_ser = dill.dumps(args)
        self.kwargs_ser = dill.dumps(kwargs)
        self._dry_objects = None
        self._activated_objects = None

    def final_call(
            self,
            f,
            ctx_ret_q,
            tune_report_q,
            checkpoint_req_q,
            checkpoint_ret_q,
            *args,
            **kwargs):

        if ctx_ret_q is None:
            raise RuntimeError("A return queue is required.")

        if tune_report_q is None:
            raise RuntimeError("A report queue is required.")

        if checkpoint_req_q is None:
            raise RuntimeError("A checkpoint request queue is required.")

        if checkpoint_ret_q is None:
            raise RuntimeError("A checkpoint return queue is required.")

        def cleanup_process(ret_val):
            # Put result in return queue
            ctx_ret_q.put(ret_val)

            # Put object updates in queue
            if len(self.update_obj_defs) > 0:
                for obj_def in self.update_obj_defs:
                    found = False
                    for obj in self._dry_objects:
                        if obj_def == obj.definition():
                            res_buf = io.BytesIO()
                            obj.save_self(res_buf)
                            res_buf.seek(0)
                            ctx_ret_q.put(res_buf.read())
                            found = True
                            break
                    if not found:
                        raise RuntimeError(
                            "Couldn't find an object to associate"
                            " a definition to!")

            # Wait until return queue is empty
            # (parent thread has emptied it)
            while not ctx_ret_q.empty():
                # Sleep a bit
                time.sleep(0.1)

            # Wait until tune report queue is empty
            # (parent thread has emptied it)
            while not tune_report_q.empty():
                time.sleep(0.1)

            # Wait until checkpoint request queue is empty
            # (parent thread should empty it.)
            while not checkpoint_req_q.empty():
                time.sleep(0.1)

        # Activate context
        with contexts[self.ctx_name][1](**self.context_kwargs):
            # Get list of dry_objects
            self._dry_objects = get_dry_objects(*args, **kwargs)
            # Activate unactivated objects
            self._activated_objects = activate_objects(self._dry_objects)

            try:
                res = f(
                    *args, **kwargs, tune_report_q=tune_report_q,
                    checkpoint_req_q=checkpoint_req_q,
                    checkpoint_ret_q=checkpoint_ret_q)
                cleanup_process(res)
            except KeyboardInterrupt:
                cleanup_process(None)

    def __call__(
            self, ctx_ret_q, tune_report_q=None,
            checkpoint_ret_q=None, checkpoint_req_q=None):
        # Undill function
        f = dill.loads(self.f_ser)

        # Undill args/kwargs
        args = dill.loads(self.args_ser)
        kwargs = dill.loads(self.kwargs_ser)

        self.final_call(
            f, ctx_ret_q, tune_report_q, checkpoint_req_q,
            checkpoint_ret_q, *args, **kwargs)


def tune_compute_context(
        ctx_context_type=None,
        ctx_use_existing_context=True,
        ctx_dont_create_context=False,
        ctx_update_objs=False,
        ctx_verbose=False,
        **ctx_context_kwargs):

    """
        A compute context allowing the retrieval of intermediate results
        to be sent to tune during training.
        ctx_context_type: if None, will try to guess
        ctx_use_existing_context: if True, will try to run in
            the currently active context without a subprocess
        ctx_dont_create_context: if True, will avoid trying to create
            a context, instead relying on the existing context.
        ctx_update_objs: if True, will serialize objects after method is run
            and update local objects by running load_object with
            the serialized results
    """

    def _func_dec(f):
        if hasattr(f, '__dry_context_wrapped__'):
            # Don't wrap a function twice
            if f.__dry_context_wrapped__:
                return f

        nonlocal ctx_context_kwargs
        if ctx_context_kwargs is None:
            ctx_context_kwargs = {}

        @functools.wraps(f)
        def wrapped_func(
                *args,
                call_context_type=None,
                call_update_objs=None,
                call_update_skiplist=None,
                call_context_kwargs=None,
                call_verbose=None,
                **kwargs):
            """
                call_*: call version of the ctx variables which take
                    precedence.
                call_update_skiplist: A list of objects to skip
                    when updating objects.
            """
            from ray import tune

            verbose = ctx_verbose
            if call_verbose is not None:
                verbose = call_verbose

            ctx_name = ctx_context_type
            if call_context_type is not None:
                ctx_name = call_context_type

            if ctx_name is None:
                # Determine context type
                ctxs = []
                for obj in get_dry_objects(*args, **kwargs):
                    ctxs.append(obj.dry_compute_context())
                ctx_name = consolidate_contexts(ctxs)

            # Execute the method in another thread.
            # get a list of objects we will need to update.
            update_objs = False
            if call_update_objs is not None:
                update_objs = call_update_objs
            elif ctx_update_objs:
                update_objs = True

            # record which objects to update
            update_objs_list = []
            update_obj_defs = []
            if update_objs:
                # build update skiplist
                if call_update_skiplist is None:
                    update_skiplist = []
                else:
                    update_skiplist = call_update_skiplist

                dry_objects = get_dry_objects(*args, **kwargs)

                # Get list of dry objects
                update_objs_list = list(filter(
                    lambda o: o not in update_skiplist,
                    dry_objects))

                # Translate direct list into definitions
                update_obj_defs = list(map(
                    lambda o: o.definition(),
                    update_objs_list))

            # Prepare context args
            if call_context_kwargs is None:
                call_context_kwargs = {}

            context_kwargs = copy.copy(ctx_context_kwargs)
            context_kwargs.update(call_context_kwargs)

            # Create return queue
            ctx_ret_q = mp_ctx.Queue()

            executor = tune_process_executor(
                f=f, ctx_name=ctx_name,
                context_kwargs=context_kwargs,
                update_obj_defs=update_obj_defs,
                args=args,
                kwargs=kwargs)

            # Create tune report queue
            tune_report_q = mp_ctx.Queue()

            # Create checkpoint queues
            checkpoint_req_q = mp_ctx.Queue()
            checkpoint_ret_q = mp_ctx.Queue()

            try:
                from pytest_cov.embed import cleanup_on_sigterm
            except ImportError:
                pass
            else:
                cleanup_on_sigterm()

            # run function
            p = Process(
                target=executor, args=[ctx_ret_q],
                kwargs={
                    'tune_report_q': tune_report_q,
                    'checkpoint_req_q': checkpoint_req_q,
                    'checkpoint_ret_q': checkpoint_ret_q})
            p.start()

            if verbose:
                print(f"Started context isolation process, pid: {p.pid}")

            # Check loop
            queue_results = []

            # Num elements which need to be fetched before joining
            num_to_fetch = 1
            num_to_fetch += len(update_objs_list)

            def check_queues():
                nonlocal tune_report_q
                nonlocal ctx_ret_q
                nonlocal checkpoint_req_q
                nonlocal checkpoint_ret_q
                while len(queue_results) < num_to_fetch or \
                        not tune_report_q.empty():
                    # Check whether the queue is empty
                    if not ctx_ret_q.empty():
                        # Get result
                        queue_results.append(ctx_ret_q.get())
                    elif not tune_report_q.empty():
                        report_dict = tune_report_q.get()
                        if type(report_dict) is not dict:
                            raise ValueError(
                                "Value in tune report queue was "
                                "not a dictionary!")
                        # Report result to tune
                        tune.report(**report_dict)
                    elif not checkpoint_req_q.empty():
                        step_req = checkpoint_req_q.get()
                        with tune.checkpoint_dir(step=step_req) as d:
                            checkpoint_ret_q.put(d)
                    else:
                        break

            while p.is_alive():
                # check for exception
                if p.exception is not None:
                    e, tb = p.exception
                    print(
                        "Exception encountered in context "
                        f"thread! pid: {p.pid}")
                    print(tb)
                    # rejoin thread
                    p.join()
                    # Reraise
                    raise e

                # Get results, we need to do this before the join
                # to prevent deadlock from large data being shuttled
                # through the queue.
                check_queues()

                # Sleep a bit
                time.sleep(0.1)

            if p.exception is not None:
                e, tb = p.exception
                print("Exception encountered in context "
                      f"thread! pid: {p.pid}")
                print(tb)
                # rejoin thread
                p.join()
                # Reraise
                raise e

            # Check queue again
            check_queues()

            # join the thread
            p.join()

            if len(queue_results) < num_to_fetch:
                raise RuntimeError("Didn't get all expected results!")

            # retrieve return value
            retval = queue_results.pop(0)

            # Update dry objects
            for obj in update_objs_list:
                obj_buf = queue_results.pop(0)
                zf = zipfile.ZipFile(io.BytesIO(obj_buf), mode='r')
                obj.load_object(zf)

            # Close thread
            p.close()

            # Close and Delete the temporary queue
            ctx_ret_q.close()
            del ctx_ret_q

            # Close and Delete the tune report queue
            tune_report_q.close()
            del tune_report_q

            # Close and Delete the checkpoint queues
            checkpoint_req_q.close()
            del checkpoint_req_q

            # Wait until the checkpoint return queue is empty
            while not checkpoint_ret_q.empty():
                time.sleep(0.1)

            checkpoint_ret_q.close()
            del checkpoint_ret_q

            if verbose:
                print(f"Ended context isolation process, pid: {p.pid}")

            return retval

        wrapped_func.__dry_context_wrapped__ = True

        # Return wrapped function
        return wrapped_func

    return _func_dec



import io
from io import BufferedIOBase
import tempfile
import zipfile


class FileIntermediary(BufferedIOBase):
    def __init__(self, mem_mode=False):
        self.mem_mode = mem_mode
        if mem_mode:
            self.tmp_file = io.BytesIO()
        else:
            self.tmp_file = tempfile.NamedTemporaryFile(mode='w+b')

    # Implement IOBase, and Buffered IOBase members
    def close(self):
        self.tmp_file.close()

    @property
    def closed(self):
        return self.tmp_file.closed

    def fileno(self):
        return self.tmp_file.fileno()

    def flush(self):
        return self.tmp_file.flush()

    def isatty(self):
        return self.tmp_file.isatty()

    def readable(self):
        return self.tmp_file.readable()

    def readline(self, size=-1):
        return self.tmp_file.readline(size)

    def readlines(self, hint=-1):
        return self.tmp_file.readlines(hint)

    def seek(self, offset, whence=io.SEEK_SET):
        return self.tmp_file.seek(offset, whence)

    def seekable(self):
        return self.tmp_file.seekable()

    def tell(self):
        return self.tmp_file.tell()

    def truncate(self, size=None):
        return self.tmp_file.truncate(size)

    def writable(self):
        return self.tmp_file.writable()

    def writelines(self, lines):
        return self.tmp_file.writelines(lines)

    def read(self, size=-1):
        return self.tmp_file.read(size)

    def read1(self, size=-1):
        return self.tmp_file.read1(size)

    def readinto(self, b):
        return self.tmp_file.readinto(b)

    def readinto1(self, b):
        return self.tmp_file.readinto1(b)

    def write(self, b):
        return self.tmp_file.write(b)

    def size(self):
        old_pos = self.tell()
        self.seek(0)
        self.seek(0, 2)
        size = self.tell()
        self.seek(old_pos)
        return size

    def is_empty(self):
        size = self.size()
        empty = False
        if size != 0:
            old_pos = self.tell()
            self.seek(0)
            with zipfile.ZipFile(
                    self, mode='r') as zf:
                if len(zf.namelist()) == 0:
                    empty = True
            self.seek(old_pos)
        return empty

    # Implement with statement magic methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    # My methods
    def write_to_file(self, file):
        # Save current file position
        cur_pos = self.tell()

        # Flush to disk, and Seek to beginning
        self.flush()
        self.seek(0)

        # Read current file content into file
        if type(file) is str:
            file = open(file, 'wb')

        self.seek(0)
        file.write(self.read())

        # Restore position
        self.seek(cur_pos)
